import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import numpy as np


def bland_altman_plot(path, data1: np.ndarray, data2: np.ndarray, variable_name):
    mean = np.mean([data1, data2], axis=0)
    diff = data1 - data2
    md = np.mean(diff)  # mean of the difference
    sd = np.std(diff, axis=0)  # standard deviation of the difference

    fig = plt.figure()
    ax = fig.add_subplot(111)  # create an axes instance (nrows=ncols=index)
    ax.scatter(mean, diff, s=5, color='black')

    ax.set_title('Bland-Altman')
    ax.set_ylabel('$\Delta${}'.format(variable_name))
    ax.set_xlabel(variable_name)

    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)

    plt.axhline(md, color='gray', linestyle='--')
    plt.axhline(md + 1.96 * sd, color='gray', linestyle='--')
    plt.axhline(md - 1.96 * sd, color='gray', linestyle='--')

    # https://stackoverflow.com/questions/43675355/python-text-to-second-y-axis?noredirect=1&lq=1
    _, x = plt.gca().get_xlim()
    plt.text(x, md + 1.96 * sd, '+1.96 SD', ha='left', va='bottom')
    plt.text(x, md + 1.96 * sd, '{:.3f}'.format(md + 1.96 * sd), ha='left', va='top')

    plt.text(x, md - 1.96 * sd, '-1.96 SD', ha='left', va='bottom')
    plt.text(x, md - 1.96 * sd, '{:.3f}'.format(md - 1.96 * sd), ha='left', va='top')

    plt.text(x, md, 'Mean', ha='left', va='bottom')
    plt.text(x, md, '{:.3f}'.format(md), ha='left', va='top')

    fig.subplots_adjust(right=0.89)  # adjust slightly such that "+1.96 SD" is not cut off

    plt.savefig(path)
    plt.close()


def box_and_whisker_plot(self, file_path: str, data, title: str, x_label: str, y_label: str,
                         min_: float = None, max_: float = None):
    fig = plt.figure(figsize=plt.rcParams["figure.figsize"][::-1])  # figsize defaults to (width, height)=(6.4, 4.8)
    # for boxplots, we want the ratio to be inversed
    ax = fig.add_subplot(111)  # create an axes instance (nrows=ncols=index)
    bp = ax.boxplot(data, widths=0.6)
    self.set_box_format(bp)

    ax.set_title(title)
    ax.set_ylabel(y_label)
    if x_label is not None:
        ax.set_xlabel(x_label)

    # remove frame
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    # thicken frame
    ax.spines['left'].set_linewidth(2)
    ax.spines['bottom'].set_linewidth(2)

    # adjust min and max if provided
    if min_ is not None or max_ is not None:
        min_original, max_original = ax.get_ylim()
        min_ = min_ if min_ is not None and min_ < min_original else min_original
        max_ = max_ if max_ is not None and max_ > max_original else max_original
        ax.set_ylim(min_, max_)

    plt.savefig(file_path)
    plt.close()


def correlation_plot(path, x_data, y_data, x_label, y_label, title,
                     with_regression_line: bool = True, with_confidence_interval: bool = True,
                     with_abline: bool = True):

    # Create the plot object
    _, ax = plt.subplots()

    # Plot the data, set the size (s), color and transparency (alpha)
    # of the points
    ax.scatter(x_data, y_data, s=5, color='black')

    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)

    # add abline
    if with_abline:
        # import matplotlib.lines as mlines
        # line = mlines.Line2D([0, 1], [0, 1], color='red')
        # transform = ax.transAxes
        # line.set_transform(transform)
        # ax.add_line(line)

        x = np.linspace(*ax.get_xlim())
        ax.plot(x, x, color='black', linestyle='dashed')

    z = np.polyfit(x_data, y_data, 1)

    if with_regression_line:
        p = np.poly1d(z)
        fit = p(x_data)

        # get the coordinates for the fit curve
        c_y = [np.min(fit), np.max(fit)]
        c_x = [np.min(x_data), np.max(x_data)]

        # plot line of best fit
        ax.plot(c_x, c_y, color='black', label='Regression line')

    # see https://stackoverflow.com/questions/27164114/show-confidence-limits-and-prediction-limits-in-scatter-plot

    # if with_confidence_interval:
    #     # ax.fill_between(t, lower_bound, upper_bound, facecolor='yellow', alpha=0.5,
    #     #                 label='1 sigma range')
    #
    #     # predict y values of origional data using the fit
    #     p_y = z[0] * x_data + z[1]
    #
    #     # calculate the y-error (residuals)
    #     y_err = y_data - p_y
    #
    #     # create series of new test x-values to predict for
    #     p_x = np.arange(np.min(x_data), np.max(x_data) + 1, 1)
    #
    #     # now calculate confidence intervals for new test x-series
    #     mean_x = np.mean(x_data)  # mean of x
    #     n = len(x_data)  # number of samples in original fit
    #     t = 1.98  # appropriate t value (where n=9, two tailed 95%)
    #     s_err = np.sum(np.power(y_err, 2))  # sum of the squares of the residuals
    #
    #     confs = t * np.sqrt((s_err / (n - 2)) * (1.0 / n + (np.power((p_x - mean_x), 2) /
    #                                                         ((np.sum(np.power(x_data, 2))) - n * (
    #                                                             np.power(mean_x, 2))))))
    #
    #     # now predict y based on test x-values
    #     p_y = z[0] * p_x + z[0]
    #
    #     # get lower and upper confidence limits based on predicted y and confidence intervals
    #     lower = p_y - abs(confs)
    #     upper = p_y + abs(confs)
    #
    #     ax.plot(p_x, lower, 'b--', label='Lower confidence limit (95%)')
    #     ax.plot(p_x, upper, 'b--', label='Upper confidence limit (95%)')

    # Label the axes and provide a title
    ax.set_title(title)
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)

    min_ = min(x_data.min(), y_data.min())
    max_ = max(x_data.max(), y_data.max())

    ax.set_xlim([min_, max_])
    ax.set_ylim([min_, max_])

    plt.savefig(path)
    plt.close()


def residual_plot(path, predicted, reference, x_label, y_label, title):

    # Create the plot object
    _, ax = plt.subplots()

    # Plot the data, set the size (s), color and transparency (alpha)
    # of the points
    residuals = reference - predicted

    ax.scatter(predicted, residuals, s=5, color='black')

    # Label the axes and provide a title
    ax.set_title(title)
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)

    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)

    # min_ = min(x_data.min(), y_data.min())
    # max_ = max(x_data.max(), y_data.max())
    #
    # ax.set_xlim([min_, max_])
    # ax.set_ylim([min_, max_])

    plt.savefig(path)
    plt.close()



try:
    import tensorflow
except ImportError:
    raise ImportError('Module "tensorflow" not found. Install "tensorflow<2.0.0" to use the deeplearning module.')

try:
    import torch
except ImportError:
    raise ImportError('Module "torch" not found. Install "torch==1.3.1" to use the deeplearning module.')

try:
    import tensorboardX
except ImportError:
    raise ImportError('Module "tensorboardX" not found. Install "tensorboardX==1.9" to use the deeplearning module.')



import abc
import logging
import math
import os
import random
import time

import numpy as np
import tensorflow as tf
import torch

import pymia.data.assembler as asmbl
import pymia.deeplearning.config as cfg
import pymia.deeplearning.data_handler as hdlr
import pymia.deeplearning.logging as log
import pymia.deeplearning.model as mdl


class Trainer(abc.ABC):

    def __init__(self, data_handler: hdlr.DataHandler, logger: log.Logger, config: cfg.DeepLearningConfiguration,
                 model: mdl.Model):
        """Initializes a new instance of the Trainer class.

        The subclasses need to implement following methods:

         - train_batch
         - validate_batch
         - validate_on_subject
         - set_seed
         - init_subject_assembler
         - _check_and_load_if_model_exists

        Args:
            data_handler: A data handler for the training and validation datasets.
            logger: A logger, which logs the training process.
            config: A configuration with training parameters.
            model: The model to train.
        """
        self.epochs = config.epochs
        self.current_epoch = 1  # the current epoch
        self.current_step = 0  # the current step is equal to the amount of trained batches (TensorFlow notation)
        self.epoch_duration = 0  # the duration of one epoch in seconds

        self.model_dir = config.model_dir
        # note that in case of TensorFlow, the model path should be without an extension (e.g., '/your/path/model')
        # TensorFlow will then automatically append the epoch and file extensions.
        self.model_path = os.path.join(config.model_dir, config.model_file_name)
        self.best_model_path = os.path.join(config.model_dir, config.best_model_file_name)

        self.data_handler = data_handler
        self.logger = logger
        self.model = model

        self.validate_nth_epoch = config.validate_nth_epoch

        # the best model score used to save the best model and also to log the training progress
        self.best_model_score = math.inf  # the score (e.g., the value of a metric) of the best performing model
        self.best_model_score_is_positive = config.best_model_score_is_positive  # whether the best score is positive
        if self.best_model_score_is_positive:
            self.best_model_score = -math.inf
        self.best_model_score_name = config.best_model_score_name

        self.seed = config.seed  # used as initial seed in set_seed

        # logging properties
        self.log_nth_epoch = config.log_nth_epoch
        self.log_nth_batch = config.log_nth_batch
        self.log_visualization_nth_epoch = config.log_visualization_nth_epoch
        self.save_model_nth_epoch = config.save_model_nth_epoch
        self.save_validation_nth_epoch = config.save_validation_nth_epoch

    def train(self):
        """Trains a model, i.e. runs the epochs loop.

        This is the main method, which will call the other methods.
        """

        self._check_and_load_if_model_exists()
        for epoch in range(self.current_epoch, self.epochs + 1):
            self.current_epoch = epoch
            self.model.set_epoch(self.current_epoch)

            logging.info('Epoch {}: Start training'.format(self._get_current_epoch_formatted()))
            start_time = time.time()  # measure the epoch's time

            self.set_seed()
            subject_assembler = self.init_subject_assembler()  # todo: not optimal solution since it keeps everything in memory
            self.data_handler.dataset.set_extractor(self.data_handler.extractor_train)
            self.data_handler.dataset.set_transform(self.data_handler.extraction_transform_train)
            loss_sum = 0
            for batch_idx, batch in enumerate(self.data_handler.loader_train):
                prediction, loss_value = self.train_batch(batch_idx, batch)
                subject_assembler.add_batch(prediction, batch)
                loss_sum += loss_value

            # normalize loss value by the number of subjects
            loss_value_of_epoch = loss_sum / self.data_handler.no_subjects_train

            self.epoch_duration = time.time() - start_time

            logging.info('Epoch {}: Ended training in {:.5f} s'.format(self._get_current_epoch_formatted(),
                                                                       self.epoch_duration))
            logging.info('Epoch {}: Training loss of {:.5f}'.format(self._get_current_epoch_formatted(),
                                                                    loss_value_of_epoch))

            if epoch % self.log_nth_epoch == 0:
                self.logger.log_epoch(self.current_epoch, loss=loss_value_of_epoch, duration=self.epoch_duration)

            if epoch % self.log_visualization_nth_epoch == 0:
                self.logger.log_visualization(self.current_epoch)

            if epoch % self.validate_nth_epoch == 0:
                loss_validation, model_score = self.validate()

                self.logger.log_scalar('loss', loss_validation, self.current_epoch, False)
                logging.info('Epoch {}: Validation loss of {:.5f}'.format(self._get_current_epoch_formatted(),
                                                                          loss_validation))
                self.logger.log_scalar('model_score', model_score, self.current_epoch, False)
                logging.info('Epoch {}: Model score of {:.5f}'.format(self._get_current_epoch_formatted(),
                                                                      model_score))

                if (model_score > self.best_model_score and self.best_model_score_is_positive) or \
                        (model_score < self.best_model_score and not self.best_model_score_is_positive):
                    self.best_model_score = model_score
                    self.model.save(self.best_model_path, self.current_epoch, best_model_score=self.best_model_score,
                                    best_model_score_name=self.best_model_score_name)

                self.validate_on_subject(subject_assembler, True)  # also pass the test set to the validator

            if epoch % self.save_model_nth_epoch == 0:
                # note that the model saving needs to be done AFTER saving the best model (if any),
                # because for TensorFlow, saving the best model updates the best_model_score variable
                self.model.save(self.model_path, self.current_epoch)

    @abc.abstractmethod
    def train_batch(self, idx: int, batch: dict) -> (np.ndarray, float):
        """Trains a batch.

        Args:
            idx: The batch index.
            batch: The batch.

        Returns:
            The prediction and the loss value of the training.
        """
        pass

    def validate(self) -> (float, float):
        """Validates the model's performance on the validation set.

        Loops over the entire validation set, predicts the samples, and assembles them to subjects.
        Finally, validate_on_subject is called, which validates the model's performance on a subject level.

        Returns:
            (float, float): A tuple where the first element is the validation loss and the second is the model score.
        """

        subject_assembler = self.init_subject_assembler()
        self.data_handler.dataset.set_extractor(self.data_handler.extractor_valid)
        self.data_handler.dataset.set_transform(self.data_handler.extraction_transform_valid)
        loss_sum = 0
        for batch_idx, batch in enumerate(self.data_handler.loader_valid):
            prediction, loss_value = self.validate_batch(batch_idx, batch)
            # prediction = np.stack(batch['labels'], axis=0)  # use this line to verify assembling
            subject_assembler.add_batch(prediction, batch)
            loss_sum += loss_value

        # normalize loss value by the number of subjects
        loss_value_of_epoch = loss_sum / self.data_handler.no_subjects_valid

        return loss_value_of_epoch, self.validate_on_subject(subject_assembler, False)

    @abc.abstractmethod
    def validate_batch(self, idx: int, batch: dict) -> (np.ndarray, float):
        """Validates a batch.

        Args:
            idx: The batch index.
            batch: The batch.

        Returns:
            A tuple with the prediction and the loss value.
        """
        pass

    @abc.abstractmethod
    def validate_on_subject(self, subject_assembler: asmbl.Assembler, is_training: bool) -> float:
        """Validates the model's performance on the subject level.

        Args:
            subject_assembler: A subject assembler with the assembled predictions of the subjects.
            is_training: Indicates whether it are predictions on the training set or the validate set.

        Returns:
            float: The model score.
        """
        pass

    @abc.abstractmethod
    def set_seed(self):
        """Sets the seed."""
        pass

    @abc.abstractmethod
    def init_subject_assembler(self) -> asmbl.Assembler:
        """Initializes a subject assembler.

        The subject assembler is used in validate to assemble the predictions on the validation set.

        Returns:
            A subject assembler.
        """
        pass

    @abc.abstractmethod
    def _check_and_load_if_model_exists(self):
        """Checks and loads a model if it exists.

        The model should be loaded from the model_dir property. Note that this function needs to set the
        current_epoch, current_step, and best_model_score properties accordingly.
        """
        pass

    def _get_current_epoch_formatted(self) -> str:
        """Gets the current epoch formatted with leading zeros.

        Returns:
            A string with the formatted current epoch.
        """
        return str(self.current_epoch).zfill(len(str(self.epochs)))

    def _get_batch_index_formatted(self, batch_idx) -> str:
        """Gets the current batch index formatted with leading zeros.

        Returns:
            A string with the formatted batch index.
        """
        return str(batch_idx + 1).zfill(len(str(len(self.data_handler.loader_train))))


class TensorFlowTrainer(Trainer, abc.ABC):

    def __init__(self, data_handler: hdlr.DataHandler, logger: log.Logger, config: cfg.DeepLearningConfiguration,
                 model: mdl.TensorFlowModel, session: tf.Session):
        """Initializes a new instance of the TensorFlowTrainer class.

        The subclasses need to implement following methods:

         - validate_on_subject
         - init_subject_assembler

        Args:
            data_handler: A data handler for the training and validation datasets.
            logger: A logger, which logs the training process.
            config: A configuration with training parameters.
            model: The model to train.
            session: A TensorFlow session.
        """
        super().__init__(data_handler, logger, config, model)

        self.session = session
        self.model = model  # set by base class too but done here such that IDE knows it is of type TensorFlowModel

        # init TensorFlow
        global_initializer_op = tf.global_variables_initializer()
        local_initializer_op = tf.local_variables_initializer()  # e.g. for tf.metrics variables
        self.session.run(global_initializer_op)
        self.session.run(local_initializer_op)

    def batch_to_feed_dict(self, batch: dict, is_training: bool):
        """Generates the TensorFlow feed dictionary.

        This basic implementation adds x, y, and is_training to the feed dictionary. Override this method to add further
        data to the feed dictionary.

        Args:
            batch: The batch from the data loader.
            is_training: Indicates whether it is the training or validation / testing phase.

        Returns:
            The feed dictionary.
        """
        feed_dict = {self.model.x_placeholder: np.stack(batch['images'], axis=0),
                     self.model.y_placeholder: np.stack(batch['labels'], axis=0),
                     self.model.is_training_placeholder: is_training}

        return feed_dict

    def train_batch(self, idx, batch: dict):
        feed_dict = self.batch_to_feed_dict(batch, True)

        prediction, _, loss_val = self.session.run([self.model.network, self.model.optimizer, self.model.loss],
                                                   feed_dict=feed_dict)

        if idx % self.log_nth_batch == 0:
            self.logger.log_batch(self.current_step, feed_dict=feed_dict)

            logging.info('Epoch {}, batch {}/{:d}: loss={:5f}'
                         .format(self._get_current_epoch_formatted(),
                                 self._get_batch_index_formatted(idx),
                                 len(self.data_handler.loader_train),
                                 loss_val))

        self.current_step += 1

        return prediction, loss_val * len(batch['images'])

    def validate_batch(self, idx: int, batch: dict) -> (np.ndarray, float):
        """Validates a batch.

        Args:
            idx: The batch index.
            batch: The batch.

        Returns:
            A tuple with the prediction and the loss value.
        """
        feed_dict = self.batch_to_feed_dict(batch, False)
        prediction, loss_val = self.session.run([self.model.network, self.model.loss], feed_dict=feed_dict)
        return prediction, loss_val * len(batch['images'])

    def set_seed(self):
        """Sets the seed depending of the current epoch.

        The seed is updated at the beginning of every epoch to ensure reproducible experiments.
        """
        seed = self.seed + self.current_epoch
        logging.info('Epoch {}: Set seed to {}'.format(self._get_current_epoch_formatted(), seed))
        random.seed(seed)
        np.random.seed(seed)
        tf.set_random_seed(seed)
        torch.manual_seed(seed)  # because pymia.data depends on torch

    def _check_and_load_if_model_exists(self):
        if self.model.load(self.model_dir):
            self.current_epoch = self.model.epoch.eval() + 1  # we save models always AFTER we finished an epoch,
            # now we enter the next epoch
            self.current_step = self.model.global_step.eval()  # global step is incremented AFTER we have seen a batch
            self.best_model_score = self.model.best_model_score.eval()
        else:
            self.current_epoch = 1
            self.current_step = 0

    @abc.abstractmethod
    def init_subject_assembler(self) -> asmbl.Assembler:
        raise NotImplementedError('init_subject_assembler')

    @abc.abstractmethod
    def validate_on_subject(self, subject_assembler: asmbl.Assembler, is_training: bool):
        raise NotImplementedError('validate_on_subject')


class TorchTrainer(Trainer, abc.ABC):

    def __init__(self, data_handler: hdlr.DataHandler, logger: log.Logger, config: cfg.DeepLearningConfiguration,
                 model: mdl.TorchModel):
        """Initializes a new instance of the TorchTrainer class.

        The subclasses need to implement following methods:

         - validate_on_subject
         - init_subject_assembler

        Args:
            data_handler: A data handler for the training and validation datasets.
            logger: A logger, which logs the training process.
            config: A configuration with training parameters.
            model: The model to train.
            cudnn_deterministic: Set CUDA to be deterministic if True, otherwise results might not be fully reproducible.
        """
        super().__init__(data_handler, logger, config, model)

        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        self.model = model  # only required for IntellISense
        self.model.device = self.device
        self.model.network.to(self.device)
        self.cudnn_deterministic = config.cudnn_determinism

    def train_batch(self, idx, batch: dict):
        images = self._get_x(batch)
        labels = self._get_y(batch)

        self.model.network.train()
        self.model.optimizer.zero_grad()

        prediction = self.model.inference(images)
        loss_val = self.model.loss_function(prediction, labels, batch=batch)
        loss_val.backward()
        self.model.optimize()

        if idx % self.log_nth_batch == 0:
            logging.info('Epoch {}, batch {}/{:d}: loss={:5f}'
                         .format(self._get_current_epoch_formatted(),
                                 self._get_batch_index_formatted(idx),
                                 len(self.data_handler.loader_train),
                                 loss_val.item()))

        self.current_step += 1

        return prediction.cpu().detach().numpy(), loss_val.item() * batch['images'].size()[0]

    def validate_batch(self, idx: int, batch: dict) -> (np.ndarray, float):
        """Evaluates a batch.

        Args:
            idx: The batch index.
            batch: The batch.
        """

        images = self._get_x(batch)
        labels = self._get_y(batch)

        self.model.network.eval()

        with torch.no_grad():
            prediction = self.model.inference(images)
            loss_val = self.model.loss_function(prediction, labels, batch=batch)

        return prediction.cpu().numpy(), loss_val.item() * batch['images'].size()[0]

    def set_seed(self):
        """Sets the seed depending of the current epoch."""
        seed = self.seed + self.current_epoch
        logging.info('Epoch {}: Set seed to {}'.format(self._get_current_epoch_formatted(), seed))
        random.seed(seed)
        np.random.seed(seed)
        torch.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
        if self.cudnn_deterministic:
            torch.backends.cudnn.deterministic = True
            torch.backends.cudnn.benchmark = False

    def _check_and_load_if_model_exists(self):
        if self.model.load(self.model_dir):
            self.current_epoch = self.model.epoch + 1  # we save models always AFTER we finished an epoch,
            # now we enter the next epoch
            self.current_step = self.model.global_step  # global step is incremented AFTER we have seen a batch
            self.best_model_score = self.model.best_model_score
        else:
            self.current_epoch = 1
            self.current_step = 0

    def _get_x(self, batch):
        return batch['images'].to(self.device)

    def _get_y(self, batch):
        return batch['labels'].to(self.device)

    @abc.abstractmethod
    def init_subject_assembler(self) -> asmbl.Assembler:
        raise NotImplementedError('init_subject_assembler')

    @abc.abstractmethod
    def validate_on_subject(self, subject_assembler: asmbl.Assembler, is_training: bool):
        raise NotImplementedError('evaluate_on_subject')



import abc
import typing as t

import numpy as np
import SimpleITK as sitk

import pymia.data.conversion as conv
import pymia.data.subjectfile as subj


class Load(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def __call__(self, file_name: str, id_: str, category: str, subject_id: str) \
            -> t.Tuple[np.ndarray, t.Union[conv.ImageProperties, None]]:
        pass


class LoadDefault(Load):

    def __call__(self, file_name: str, id_: str, category: str, subject_id: str) -> \
            t.Tuple[np.ndarray, t.Union[conv.ImageProperties, None]]:
        img = sitk.ReadImage(file_name)
        return sitk.GetArrayFromImage(img), conv.ImageProperties(img)






"""This module contains global definitions for the :mod:`pymia.data` package.
"""

NAMES_PLACEHOLDER = 'meta/names/{}_names'
# NAMES_IMAGE = 'meta/names/image_names'
# NAMES_LABEL = 'meta/names/label_names'
# NAMES_SUPPL = 'meta/names/supplementary_names'

INFO_SHAPE = 'meta/info/shapes'
INFO_ORIGIN = 'meta/info/origins'
INFO_DIRECTION = 'meta/info/directions'
INFO_SPACING = 'meta/info/spacing'

FILES_PLACEHOLDER = 'meta/files/{}_files'
# FILES_IMAGE = 'meta/files/image_files'
# FILES_LABEL = 'meta/files/label_files'
# FILES_SUPPL = 'meta/files/supplementary_files'
FILES_ROOT = 'meta/files/file_root'

SUBJECT = 'meta/subjects'

# DATA = 'data'
DATA_PLACEHOLDER = 'data/{}'
# DATA_IMAGE = '{}/images'.format(DATA)
# DATA_LABEL = '{}/labels'.format(DATA)

# keys for a batch dictionary
KEY_FILE_ROOT = 'file_root'
KEY_IMAGES = 'images'
KEY_LABELS = 'labels'
KEY_PROPERTIES = 'properties'
KEY_SHAPE = 'shape'
KEY_SUBJECT = 'subject'
KEY_SUBJECT_INDEX = 'subject_index'



import abc

import pymia.config.configuration as cfg


class DeepLearningConfiguration(cfg.ConfigurationBase, abc.ABC):
    """Represents a basic deep learning configuration."""

    def __init__(self):
        """Initializes a new instance of the DeepLearningConfiguration class.

        Args:
            config (dict): A dictionary representing the configuration.
        """
        self.database_file = ''
        self.result_dir = ''
        self.model_dir = ''
        self.model_file_name = 'model'  # to save the model with this file name
        self.best_model_file_name = 'model-best'

        self.seed = 42  # initial seed during training
        self.cudnn_determinism = False  # whether to enable CuDNN determinism or not

        # training configuration
        self.epochs = 200  # number of epochs
        self.batch_size_training = 4
        self.batch_size_testing = 4  # is also used for evaluation

        self.validate_nth_epoch = 1  # validate the performance each nth epoch

        self.best_model_score_is_positive = True  # whether the best model score is positive or negative
        self.best_model_score_name = 'unknown'

        # logging configuration
        self.log_nth_epoch = 1  # log each nth epoch
        self.log_nth_batch = 5  # log each nth batch
        self.log_visualization_nth_epoch = 1  # log the visualization each nth epoch
        self.save_model_nth_epoch = 1  # save the model each nth epoch
        self.save_validation_nth_epoch = 5  # save the validation results each nth epoch



"""This module holds classes related to batch feeding together with the :py:class`DataLoader`."""
import collections

import numpy as np
import torch.utils.data.dataloader as loader


class CollateBase:
    """Collate base class for merging a list of samples to a batch."""

    def __init__(self, entries=('images', 'labels')):
        """Initializes a new instance of the CollateBase class.

        Args:
            entries (:obj:`list` of :obj:`str`): The entries to merge for batch feeding to the GPU.
        """
        self.entries = entries

    def __call__(self, batch) -> dict:
        raise NotImplementedError('Derived class needs to implement __call__')


class TensorFlowCollate(CollateBase):

    def __call__(self, batch) -> dict:
        """Merges the `entries` to numpy arrays and the remaining entries in `batch` to a tuple.

        Args:
            batch (:obj:`list` of :obj:`dict`): A list of batch entries where len(batch) is the batch size.

        Returns:
            The merged batch, i.e. a dict of numpy arrays and tuples.

        Examples:
            Use the class during the instantiation of a DataLoader:

            >>> import pymia.data.extraction as pymia_extr
            >>> collate = TensorFlowCollate(entries=('images', 'labels'))
            >>> loader = pymia_extr.DataLoader(dataset=None,collate_fn=collate)  # specify an appropriate dataset
        """
        new_batch = {}
        for key in batch[0]:
            if key in self.entries:
                new_batch[key] = np.array([d[key] for d in batch])
            else:
                new_batch[key] = tuple([d[key] for d in batch])
        return new_batch


class TorchCollate(CollateBase):

    def __call__(self, batch) -> dict:
        """Merges the `entries` to Torch Tensors and the remaining entries in `batch` to a tuple.

        Args:
            batch (:obj:`list` of :obj:`dict`): A list of batch entries where len(batch) is the batch size.

        Returns:
            The merged batch, i.e. a dict of Torch Tensors and tuples.

        Examples:
            Use the class during the instantiation of a DataLoader:

            >>> import pymia.data.extraction as pymia_extr
            >>> collate = TensorFlowCollate(entries=('images', 'labels'))
            >>> loader = pymia_extr.DataLoader(dataset=None,collate_fn=collate)  # specify an appropriate dataset
        """
        if not isinstance(batch[0], collections.Mapping):
            raise TypeError('only mapping type allowed. Found {}'.format(type(batch[0])))

        new_batch = {}
        for key in batch[0]:
            if key in self.entries:
                new_batch[key] = loader.default_collate([d[key] for d in batch])
            else:
                new_batch[key] = [d[key] for d in batch]
        return new_batch



import abc


class DataHandler(abc.ABC):
    """Represents an abstract data handler for the pymia data package."""

    def __init__(self):
        self.dataset = None  # pymia_extr.ParameterizableDataset

        self.extractor_train = None  # pymia_extr.Extractor
        self.extractor_valid = None  # pymia_extr.Extractor
        self.extractor_test = None  # pymia_extr.Extractor

        self.extraction_transform_train = None  # pymia_tfm.Transform
        self.extraction_transform_valid = None  # pymia_tfm.Transform
        self.extraction_transform_test = None  # pymia_tfm.Transform

        self.loader_train = None  # torch.utils.data.DataLoader
        self.loader_valid = None  # torch.utils.data.DataLoader
        self.loader_test = None  # torch.utils.data.DataLoader

        self.no_subjects_train = 0
        self.no_subjects_valid = 0
        self.no_subjects_test = 0



import abc
import math
import warnings

import numpy as np
import SimpleITK as sitk

from .base import (IConfusionMatrixMetric, IDistanceMetric, ISimpleITKImageMetric, INumpyArrayMetric,
                   NotComputableMetricWarning)


class AreaMetric(ISimpleITKImageMetric):
    """Represents an area metric."""

    def __init__(self, metric: str = 'AREA'):
        """Initializes a new instance of the AreaMetric class.

        Args:
            metric (str): The identification string of the metric.
        """
        super().__init__(metric)

    @abc.abstractmethod
    def calculate(self):
        """Calculates the metric."""
        raise NotImplementedError

    @staticmethod
    def _calculate_area(image: sitk.Image, slice_number: int = -1) -> float:
        """Calculates the area of a slice in a label image.

        Args:
            image (sitk.Image): The 3-D label image.
            slice_number (int): The slice number to calculate the area.
                Defaults to -1, which will calculate the area on the intermediate slice.
        """

        img_arr = sitk.GetArrayFromImage(image)

        if slice_number == -1:
            slice_number = int(img_arr.shape[0] / 2)  # use the intermediate slice

        return img_arr[slice_number, ...].sum() * image.GetSpacing()[0] * image.GetSpacing()[1]


class VolumeMetric(ISimpleITKImageMetric):
    """Represents a volume metric."""

    def __init__(self, metric: str = 'VOL'):
        """Initializes a new instance of the VolumeMetric class.

        Args:
            metric (str): The identification string of the metric.
        """
        super().__init__(metric)

    @abc.abstractmethod
    def calculate(self):
        """Calculates the metric."""
        raise NotImplementedError

    @staticmethod
    def _calculate_volume(image: sitk.Image) -> float:
        """Calculates the volume of a label image.

        Args:
            image (sitk.Image): The 3-D label image.
        """

        voxel_volume = np.prod(image.GetSpacing())
        number_of_voxels = sitk.GetArrayFromImage(image).sum()

        return number_of_voxels * voxel_volume


class Accuracy(IConfusionMatrixMetric):
    """Represents an accuracy metric."""

    def __init__(self, metric: str = 'ACURCY'):
        """Initializes a new instance of the Accuracy class.

        Args:
            metric (str): The identification string of the metric.
        """
        super().__init__(metric)

    def calculate(self):
        """Calculates the accuracy."""

        sum_ = self.confusion_matrix.tp + self.confusion_matrix.tn + self.confusion_matrix.fp + self.confusion_matrix.fn

        if sum_ != 0:
            return (self.confusion_matrix.tp + self.confusion_matrix.tn) / sum_
        else:
            return 0


class AdjustedRandIndex(IConfusionMatrixMetric):
    """Represents an adjusted rand index metric."""

    def __init__(self, metric: str = 'ADJRIND'):
        """Initializes a new instance of the AdjustedRandIndex class.

        Args:
            metric (str): The identification string of the metric.
        """
        super().__init__(metric)

    def calculate(self):
        """Calculates the adjusted rand index."""

        tp = self.confusion_matrix.tp
        tn = self.confusion_matrix.tn
        fp = self.confusion_matrix.fp
        fn = self.confusion_matrix.fn
        n = self.confusion_matrix.n

        fp_tn = tn + fp
        tp_fn = fn + tp
        tn_fn = tn + fn
        tp_fp = fp + tp
        nis = tn_fn * tn_fn + tp_fp * tp_fp
        njs = fp_tn * fp_tn + tp_fn * tp_fn
        sum_of_squares = tp * tp + tn * tn + fp * fp + fn * fn

        a = (tp * (tp - 1) + fp * (fp - 1) + tn * (tn - 1) + fn * (fn - 1)) / 2.
        b = (njs - sum_of_squares) / 2.
        c = (nis - sum_of_squares) / 2.
        d = (n * n + sum_of_squares - nis - njs) / 2.

        x1 = a - ((a + c) * (a + b) / (a + b + c + d))
        x2 = ((a + c) + (a + b)) / 2.
        x3 = ((a + c) * (a + b)) / (a + b + c + d)
        denominator = x2 - x3

        if denominator != 0:
            return x1 / denominator
        else:
            return 0


class AreaUnderCurve(IConfusionMatrixMetric):
    """Represents an area under the curve metric."""

    def __init__(self, metric: str = 'AUC'):
        """Initializes a new instance of the AreaUnderCurve class.

        Args:
            metric (str): The identification string of the metric.
        """
        super().__init__(metric)

    def calculate(self):
        """Calculates the area under the curve."""

        specificity = self.confusion_matrix.tn / (self.confusion_matrix.tn + self.confusion_matrix.fp)

        false_positive_rate = 1 - specificity

        true_positive_rate = self.confusion_matrix.tp / (self.confusion_matrix.tp + self.confusion_matrix.fn)

        return (true_positive_rate - false_positive_rate + 1) / 2


class AverageDistance(ISimpleITKImageMetric):
    """Represents an average (Hausdorff) distance metric.

        Calculates the distance between the set of non-zero pixels of two images using the following equation:

        .. math:: AVD(A,B) = max(d(A,B), d(B,A)),

        where

        .. math:: d(A,B) = \\frac{1}{N} \\sum_{a \\in A} \\min_{b \\in B} \\lVert a - b \\rVert

        is the directed Hausdorff distance and :math:`A` and :math:`B` are the set of non-zero pixels in the images.
        """

    def __init__(self, metric: str = 'AVGDIST'):
        """Initializes a new instance of the AverageDistance class.

        Args:
            metric (str): The identification string of the metric.
        """
        super().__init__(metric)

    def calculate(self):
        """Calculates the average (Hausdorff) distance."""

        if np.count_nonzero(sitk.GetArrayFromImage(self.ground_truth)) == 0:
            warnings.warn('Unable to compute average distance due to empty label mask, returning inf',
                          NotComputableMetricWarning)
            return float('inf')
        if np.count_nonzero(sitk.GetArrayFromImage(self.segmentation)) == 0:
            warnings.warn('Unable to compute average distance due to empty segmentation mask, returning inf',
                          NotComputableMetricWarning)
            return float('inf')

        distance_filter = sitk.HausdorffDistanceImageFilter()
        distance_filter.Execute(self.ground_truth, self.segmentation)
        return distance_filter.GetAverageHausdorffDistance()


class CohenKappaMetric(IConfusionMatrixMetric):
    """Represents a Cohen's kappa coefficient metric."""

    def __init__(self, metric: str = 'KAPPA'):
        """Initializes a new instance of the CohenKappaMetric class.

        Args:
            metric (str): The identification string of the metric.
        """
        super().__init__(metric)

    def calculate(self):
        """Calculates the Cohen's kappa coefficient."""

        tp = self.confusion_matrix.tp
        tn = self.confusion_matrix.tn
        fp = self.confusion_matrix.fp
        fn = self.confusion_matrix.fn

        agreement = tp + tn
        chance0 = (tn + fn) * (tn + fp)
        chance1 = (fp + tp) * (fn + tp)
        sum_ = tn + fn + fp + tp
        chance = (chance0 + chance1) / sum_

        return (agreement - chance) / (sum_ - chance)


class DiceCoefficient(IConfusionMatrixMetric):
    """Represents a Dice coefficient metric with empty target handling, defined as:

        .. math:: \\begin{cases} 1 & \\left\\vert{y}\\right\\vert = \\left\\vert{\\hat y}\\right\\vert = 0 \\\\ Dice(y,\\hat y) & \\left\\vert{y}\\right\\vert > 0 \\\\ \\end{cases}

        where :math:`\\hat y` is the prediction and :math:`y` the target.
    """

    def __init__(self, metric: str = 'DICE'):
        """Initializes a new instance of the DiceCoefficient class.

        Args:
            metric (str): The identification string of the metric.
        """
        super().__init__(metric)

    def calculate(self):
        """Calculates the Dice coefficient."""

        if (self.confusion_matrix.tp == 0) and \
                ((self.confusion_matrix.tp + self.confusion_matrix.fp + self.confusion_matrix.fn) == 0):
            return 1.

        return 2 * self.confusion_matrix.tp / \
               (2 * self.confusion_matrix.tp + self.confusion_matrix.fp + self.confusion_matrix.fn)


class FalseNegative(IConfusionMatrixMetric):
    """Represents a false negative metric."""

    def __init__(self, metric: str = 'FN'):
        """Initializes a new instance of the FalseNegative class.

        Args:
            metric (str): The identification string of the metric.
        """
        super().__init__(metric)

    def calculate(self):
        """Calculates the false negatives."""

        return self.confusion_matrix.fn


class FalsePositive(IConfusionMatrixMetric):
    """Represents a false positive metric."""

    def __init__(self, metric: str = 'FP'):
        """Initializes a new instance of the FalsePositive class.

        Args:
            metric (str): The identification string of the metric.
        """
        super().__init__(metric)

    def calculate(self):
        """Calculates the false positives."""

        return self.confusion_matrix.fp


class Fallout(IConfusionMatrixMetric):
    """Represents a fallout (false positive rate) metric."""

    def __init__(self, metric: str = 'FALLOUT'):
        """Initializes a new instance of the Fallout class.

        Args:
            metric (str): The identification string of the metric.
        """
        super().__init__(metric)

    def calculate(self):
        """Calculates the fallout (false positive rate)."""

        specificity = self.confusion_matrix.tn / (self.confusion_matrix.tn + self.confusion_matrix.fp)
        return 1 - specificity


class FalseNegativeRate(IConfusionMatrixMetric):
    """Represents a false negative rate metric."""

    def __init__(self, metric: str = 'FNR'):
        """Initializes a new instance of the FalseNegativeRate class.

        Args:
            metric (str): The identification string of the metric.
        """
        super().__init__(metric)

    def calculate(self):
        """Calculates the false negative rate."""

        sensitivity = self.confusion_matrix.tp / (self.confusion_matrix.tp + self.confusion_matrix.fn)
        return 1 - sensitivity


class FMeasure(IConfusionMatrixMetric):
    """Represents a F-measure metric."""

    def __init__(self, metric: str = 'FMEASR'):
        """Initializes a new instance of the FMeasure class.

        Args:
            metric (str): The identification string of the metric.
        """
        super().__init__(metric)

    def calculate(self):
        """Calculates the F1 measure."""

        beta = 1  # or 0.5 or 2 can also calculate F2 or F0.5 measure

        beta_squared = beta * beta
        precision = Precision()
        precision.confusion_matrix = self.confusion_matrix
        precision = precision.calculate()
        recall = Recall()
        recall.confusion_matrix = self.confusion_matrix
        recall = recall.calculate()

        denominator = beta_squared * precision + recall

        if denominator != 0:
            return (1 + beta_squared) * ((precision * recall) / denominator)
        else:
            return 0


class GlobalConsistencyError(IConfusionMatrixMetric):
    """Represents a global consistency error metric.

    Implementation based on Martin 2001.
    """

    def __init__(self, metric: str = 'GCOERR'):
        """Initializes a new instance of the GlobalConsistencyError class.

        Args:
            metric (str): The identification string of the metric.
        """
        super().__init__(metric)

    def calculate(self):
        """Calculates the global consistency error."""

        tp = self.confusion_matrix.tp
        tn = self.confusion_matrix.tn
        fp = self.confusion_matrix.fp
        fn = self.confusion_matrix.fn

        n = tp + tn + fp + fn
        e1 = (fn * (fn + 2 * tp) / (tp + fn) + fp * (fp + 2 * tn) / (tn + fp)) / n
        e2 = (fp * (fp + 2 * tp) / (tp + fp) + fn * (fn + 2 * tn) / (tn + fn)) / n

        return min(e1, e2)


class GroundTruthArea(AreaMetric):
    """Represents a ground truth area metric."""

    def __init__(self, slice_number: int = -1, metric: str = 'GTAREA'):
        """Initializes a new instance of the GroundTruthArea class.

        Args:
            slice_number (int): The slice number to calculate the area.
                Defaults to -1, which will calculate the area on the intermediate slice.
            metric (str): The identification string of the metric.
        """
        super().__init__(metric)
        self.slice_number = slice_number

    def calculate(self):
        """Calculates the ground truth area on a specified slice in mm2."""

        return self._calculate_area(self.ground_truth, self.slice_number)


class GroundTruthVolume(VolumeMetric):
    """Represents a ground truth volume metric."""

    def __init__(self, metric: str = 'GTVOL'):
        """Initializes a new instance of the GroundTruthVolume class.

        Args:
            metric (str): The identification string of the metric.
        """
        super().__init__(metric)

    def calculate(self):
        """Calculates the ground truth volume in mm3."""

        return self._calculate_volume(self.ground_truth)


class HausdorffDistance(IDistanceMetric):
    """Represents a Hausdorff distance metric.

    Calculates the distance between the set of non-zero pixels of two images using the following equation:

    .. math:: H(A,B) = max(h(A,B), h(B,A)),

    where

    .. math:: h(A,B) = \\max_{a \\in A} \\min_{b \\in B} \\lVert a - b \\rVert

    is the directed Hausdorff distance and :math:`A` and :math:`B` are the set of non-zero pixels in the images.

    See Also:
        - Nikolov et al., 2018 Deep learning to achieve clinically applicable segmentation of head and neck anatomy for
            radiotherapy. `arXiv <https://arxiv.org/abs/1809.04430>`_
        - `Original implementation <https://github.com/deepmind/surface-distance>`_
    """

    def __init__(self, percentile: float = 100.0, metric: str = 'HDRFDST'):
        """Initializes a new instance of the HausdorffDistance class.

        Args:
            percentile (float): The percentile (0, 100] to compute, i.e. 100 computes the Hausdorff distance and
                95 computes the 95th Hausdorff distance.
            metric (str): The identification string of the metric.
        """
        super().__init__(metric)
        self.percentile = percentile

    def calculate(self):
        """Calculates the Hausdorff distance."""

        if self.distances.distances_gt_to_pred is not None and len(self.distances.distances_gt_to_pred) > 0:
            surfel_areas_cum_gt = np.cumsum(self.distances.surfel_areas_gt) / np.sum(self.distances.surfel_areas_gt)
            idx = np.searchsorted(surfel_areas_cum_gt, self.percentile / 100.0)
            perc_distance_gt_to_pred = self.distances.distances_gt_to_pred[
                min(idx, len(self.distances.distances_gt_to_pred) - 1)]
        else:
            warnings.warn('Unable to compute Hausdorff distance due to empty label mask, returning inf',
                          NotComputableMetricWarning)
            return float('inf')

        if self.distances.distances_pred_to_gt is not None and len(self.distances.distances_pred_to_gt) > 0:
            surfel_areas_cum_pred = (np.cumsum(self.distances.surfel_areas_pred) / 
                                     np.sum(self.distances.surfel_areas_pred))
            idx = np.searchsorted(surfel_areas_cum_pred, self.percentile / 100.0)
            perc_distance_pred_to_gt = self.distances.distances_pred_to_gt[
                min(idx, len(self.distances.distances_pred_to_gt) - 1)]
        else:
            warnings.warn('Unable to compute Hausdorff distance due to empty segmentation mask, returning inf',
                          NotComputableMetricWarning)
            return float('inf')

        return max(perc_distance_gt_to_pred, perc_distance_pred_to_gt)


class InterclassCorrelation(INumpyArrayMetric):
    """Represents a interclass correlation metric."""

    def __init__(self, metric: str = 'ICCORR'):
        """Initializes a new instance of the InterclassCorrelation class.

        Args:
            metric (str): The identification string of the metric.
        """
        super().__init__(metric)

    def calculate(self):
        """Calculates the interclass correlation."""

        gt = self.ground_truth.flatten()
        seg = self.segmentation.flatten()

        n = gt.size
        mean_gt = gt.mean()
        mean_seg = seg.mean()
        mean = (mean_gt + mean_seg) / 2

        m = (gt + seg) / 2
        ssw = np.power(gt - m, 2).sum() + np.power(seg - m, 2).sum()
        ssb = np.power(m - mean, 2).sum()

        ssw /= n
        ssb = ssb / (n - 1) * 2

        return (ssb - ssw) / (ssb + ssw)


class JaccardCoefficient(IConfusionMatrixMetric):
    """Represents a Jaccard coefficient metric."""

    def __init__(self, metric: str = 'JACRD'):
        """Initializes a new instance of the JaccardCoefficient class.

        Args:
            metric (str): The identification string of the metric.
        """
        super().__init__(metric)

    def calculate(self):
        """Calculates the Jaccard coefficient."""

        tp = self.confusion_matrix.tp
        fp = self.confusion_matrix.fp
        fn = self.confusion_matrix.fn

        return tp / (tp + fp + fn)


class MahalanobisDistance(INumpyArrayMetric):
    """Represents a Mahalanobis distance metric."""

    def __init__(self, metric: str = 'MAHLNBS'):
        """Initializes a new instance of the MahalanobisDistance class.

        Args:
            metric (str): The identification string of the metric.
        """
        super().__init__(metric)

    def calculate(self):
        """Calculates the Mahalanobis distance."""

        gt_n = np.count_nonzero(self.ground_truth)
        gt_indices = np.flip(np.where(self.ground_truth == 1), axis=0)
        gt_mean = gt_indices.mean(axis=1)
        gt_cov = np.cov(gt_indices)

        seg_n = np.count_nonzero(self.segmentation)
        seg_indices = np.flip(np.where(self.segmentation == 1), axis=0)
        seg_mean = seg_indices.mean(axis=1)
        seg_cov = np.cov(seg_indices)

        # calculate common covariance matrix
        common_cov = (gt_n * gt_cov + seg_n * seg_cov) / (gt_n + seg_n)
        common_cov_inv = np.linalg.inv(common_cov)

        mean = np.matrix(np.array(gt_mean) - np.array(seg_mean))

        return math.sqrt(mean * np.matrix(common_cov_inv) * mean.T)


class MutualInformation(IConfusionMatrixMetric):
    """Represents a mutual information metric."""

    def __init__(self, metric: str = 'MUTINF'):
        """Initializes a new instance of the MutualInformation class.

        Args:
            metric (str): The identification string of the metric.
        """
        super().__init__(metric)

    def calculate(self):
        """Calculates the mutual information."""

        tp = self.confusion_matrix.tp
        tn = self.confusion_matrix.tn
        fp = self.confusion_matrix.fp
        fn = self.confusion_matrix.fn
        n = self.confusion_matrix.n

        fn_tp = fn + tp
        fp_tp = fp + tp
        
        if fn_tp == 0 or fn_tp / n == 1 or fp_tp == 0 or fp_tp / n == 1:
            warnings.warn('Unable to compute mutual information due to log2 of 0, returning -inf',
                          NotComputableMetricWarning)
            return float('-inf')

        h1 = -((fn_tp / n) * math.log2(fn_tp / n) + (1 - fn_tp / n) * math.log2(1 - fn_tp / n))
        h2 = -((fp_tp / n) * math.log2(fp_tp / n) + (1 - fp_tp / n) * math.log2(1 - fp_tp / n))

        p00 = 1 if tn == 0 else (tn / n)
        p01 = 1 if fn == 0 else (fn / n)
        p10 = 1 if fp == 0 else (fp / n)
        p11 = 1 if tp == 0 else (tp / n)

        h12 = -((tn / n) * math.log2(p00) +
                (fn / n) * math.log2(p01) +
                (fp / n) * math.log2(p10) +
                (tp / n) * math.log2(p11))

        mi = h1 + h2 - h12
        return mi


class Precision(IConfusionMatrixMetric):
    """Represents a precision metric."""

    def __init__(self, metric: str = 'PRCISON'):
        """Initializes a new instance of the Precision class.

        Args:
            metric (str): The identification string of the metric.
        """
        super().__init__(metric)

    def calculate(self):
        """Calculates the precision."""

        sum_ = self.confusion_matrix.tp + self.confusion_matrix.fp

        if sum_ != 0:
            return self.confusion_matrix.tp / sum_
        else:
            return 0


class ProbabilisticDistance(INumpyArrayMetric):
    """Represents a probabilistic distance metric."""

    def __init__(self, metric: str = 'PROBDST'):
        """Initializes a new instance of the ProbabilisticDistance class.

        Args:
            metric (str): The identification string of the metric.
        """
        super().__init__(metric)

    def calculate(self):
        """Calculates the probabilistic distance."""

        gt = self.ground_truth.flatten().astype(np.int8)
        seg = self.segmentation.flatten().astype(np.int8)

        probability_difference = np.absolute(gt - seg).sum()
        probability_joint = (gt * seg).sum()

        if probability_joint != 0:
            return probability_difference / (2. * probability_joint)
        else:
            return -1


class RandIndex(IConfusionMatrixMetric):
    """Represents a rand index metric."""

    def __init__(self, metric: str = 'RNDIND'):
        """Initializes a new instance of the RandIndex class.

        Args:
            metric (str): The identification string of the metric.
        """
        super().__init__(metric)

    def calculate(self):
        """Calculates the rand index."""

        tp = self.confusion_matrix.tp
        tn = self.confusion_matrix.tn
        fp = self.confusion_matrix.fp
        fn = self.confusion_matrix.fn
        n = self.confusion_matrix.n

        fp_tn = tn + fp
        tp_fn = fn + tp
        tn_fn = tn + fn
        tp_fp = fp + tp
        nis = tn_fn * tn_fn + tp_fp * tp_fp
        njs = fp_tn * fp_tn + tp_fn * tp_fn
        sum_of_squares = tp * tp + tn * tn + fp * fp + fn * fn

        a = (tp * (tp - 1) + fp * (fp - 1) + tn * (tn - 1) + fn * (fn - 1)) / 2.
        b = (njs - sum_of_squares) / 2.
        c = (nis - sum_of_squares) / 2.
        d = (n * n + sum_of_squares - nis - njs) / 2.

        return (a + d) / (a + b + c + d)


class Recall(IConfusionMatrixMetric):
    """Represents a recall metric."""

    def __init__(self, metric: str = 'RECALL'):
        """Initializes a new instance of the Recall class.

        Args:
            metric (str): The identification string of the metric.
        """
        super().__init__(metric)

    def calculate(self):
        """Calculates the recall."""

        sum_ = self.confusion_matrix.tp + self.confusion_matrix.fn

        if sum_ != 0:
            return self.confusion_matrix.tp / sum_
        else:
            return 0


class SegmentationArea(AreaMetric):
    """Represents a segmentation area metric."""

    def __init__(self, slice_number: int = -1, metric: str = 'SEGAREA'):
        """Initializes a new instance of the SegmentationArea class.

        Args:
            slice_number (int): The slice number to calculate the area.
                Defaults to -1, which will calculate the area on the intermediate slice.
            metric (str): The identification string of the metric.
        """
        super().__init__(metric)
        self.slice_number = slice_number

    def calculate(self):
        """Calculates the segmentation area on a specified slice in mm2."""

        return self._calculate_area(self.segmentation, self.slice_number)


class SegmentationVolume(VolumeMetric):
    """Represents a segmentation volume metric."""

    def __init__(self, metric: str = 'SEGVOL'):
        """Initializes a new instance of the SegmentationVolume class.

        Args:
            metric (str): The identification string of the metric.
        """
        super().__init__(metric)

    def calculate(self):
        """Calculates the segmented volume in mm3."""

        return self._calculate_volume(self.segmentation)


class Sensitivity(IConfusionMatrixMetric):
    """Represents a sensitivity (true positive rate or recall) metric."""

    def __init__(self, metric: str = 'SNSVTY'):
        """Initializes a new instance of the Sensitivity class.

        Args:
            metric (str): The identification string of the metric.
        """
        super().__init__(metric)

    def calculate(self):
        """Calculates the sensitivity (true positive rate)."""

        return self.confusion_matrix.tp / (self.confusion_matrix.tp + self.confusion_matrix.fn)


class Specificity(IConfusionMatrixMetric):
    """Represents a specificity metric."""

    def __init__(self, metric: str = 'SPCFTY'):
        """Initializes a new instance of the Specificity class.

        Args:
            metric (str): The identification string of the metric.
        """
        super().__init__(metric)

    def calculate(self):
        """Calculates the specificity."""

        return self.confusion_matrix.tn / (self.confusion_matrix.tn + self.confusion_matrix.fp)


class SurfaceDiceOverlap(IDistanceMetric):
    """Represents a surface Dice coefficient overlap metric.

    See Also:
        - Nikolov et al., 2018 Deep learning to achieve clinically applicable segmentation of head and neck anatomy for
            radiotherapy. `arXiv <https://arxiv.org/abs/1809.04430>`_
        - `Original implementation <https://github.com/deepmind/surface-distance>`_
    """

    def __init__(self, tolerance: float = 1, metric: str = 'SURFDICE'):
        """Initializes a new instance of the SurfaceDiceOverlap class.

        Args:
            tolerance (float): The tolerance of the surface distance in mm.
            metric (str): The identification string of the metric.
        """
        super().__init__(metric)
        self.tolerance = tolerance

    def calculate(self):
        """Calculates the surface Dice coefficient overlap."""

        overlap_gt = np.sum(self.distances.surfel_areas_gt[self.distances.distances_gt_to_pred <= self.tolerance])
        overlap_pred = np.sum(self.distances.surfel_areas_pred[self.distances.distances_pred_to_gt <= self.tolerance])
        surface_dice = (overlap_gt + overlap_pred) / \
                       (np.sum(self.distances.surfel_areas_gt) + np.sum(self.distances.surfel_areas_pred))
        return float(surface_dice)


class SurfaceOverlap(IDistanceMetric):
    """Represents a surface overlap metric.

    Computes the overlap of the ground truth surface with the predicted surface and vice versa allowing a
    specified tolerance (maximum surface-to-surface distance that is regarded as overlapping).
    The overlapping fraction is computed by correctly taking the area of each surface element into account.

    See Also:
        - Nikolov et al., 2018 Deep learning to achieve clinically applicable segmentation of head and neck anatomy for
            radiotherapy. `arXiv <https://arxiv.org/abs/1809.04430>`_
        - `Original implementation <https://github.com/deepmind/surface-distance>`_
    """

    def __init__(self, tolerance: float = 1.0, prediction_to_label: bool = True, metric: str = 'SURFOVLP'):
        """Initializes a new instance of the SurfaceOverlap class.

        Args:
            tolerance (float): The tolerance of the surface distance in mm.
            prediction_to_label (bool): Computes the prediction to labels if `True`, otherwise the label to prediction.
            metric (str): The identification string of the metric.
        """
        super().__init__(metric)
        self.tolerance = tolerance
        self.prediction_to_label = prediction_to_label

    def calculate(self):
        """Calculates the surface overlap."""

        if self.prediction_to_label:
            return float(np.sum(self.distances.surfel_areas_pred[self.distances.distances_pred_to_gt <= self.tolerance])
                         / np.sum(self.distances.surfel_areas_pred))
        else:
            return float(np.sum(self.distances.surfel_areas_gt[self.distances.distances_gt_to_pred <= self.tolerance])
                         / np.sum(self.distances.surfel_areas_gt))


class TrueNegative(IConfusionMatrixMetric):
    """Represents a true negative metric."""

    def __init__(self, metric: str = 'TN'):
        """Initializes a new instance of the TrueNegative class.

        Args:
            metric (str): The identification string of the metric.
        """
        super().__init__(metric)

    def calculate(self):
        """Calculates the true negatives."""

        return self.confusion_matrix.tn


class TruePositive(IConfusionMatrixMetric):
    """Represents a true positive metric."""

    def __init__(self, metric: str = 'TP'):
        """Initializes a new instance of the TruePositive class.

        Args:
            metric (str): The identification string of the metric.
        """
        super().__init__(metric)

    def calculate(self):
        """Calculates the true positives."""

        return self.confusion_matrix.tp


class VariationOfInformation(IConfusionMatrixMetric):
    """Represents a variation of information metric."""

    def __init__(self, metric: str = 'VARINFO'):
        """Initializes a new instance of the VariationOfInformation class.

        Args:
            metric (str): The identification string of the metric.
        """
        super().__init__(metric)

    def calculate(self):
        """Calculates the variation of information."""

        tp = self.confusion_matrix.tp
        tn = self.confusion_matrix.tn
        fp = self.confusion_matrix.fp
        fn = self.confusion_matrix.fn
        n = self.confusion_matrix.n

        fn_tp = fn + tp
        fp_tp = fp + tp

        if fn_tp == 0 or fn_tp / n == 1 or fp_tp == 0 or fp_tp / n == 1:
            warnings.warn('Unable to compute variation of information due to log2 of 0, returning -inf',
                          NotComputableMetricWarning)
            return float('-inf')

        h1 = -((fn_tp / n) * math.log2(fn_tp / n) + (1 - fn_tp / n) * math.log2(1 - fn_tp / n))
        h2 = -((fp_tp / n) * math.log2(fp_tp / n) + (1 - fp_tp / n) * math.log2(1 - fp_tp / n))

        p00 = 1 if tn == 0 else (tn / n)
        p01 = 1 if fn == 0 else (fn / n)
        p10 = 1 if fp == 0 else (fp / n)
        p11 = 1 if tp == 0 else (tp / n)

        h12 = -((tn / n) * math.log2(p00) +
                (fn / n) * math.log2(p01) +
                (fp / n) * math.log2(p10) +
                (tp / n) * math.log2(p11))

        mi = h1 + h2 - h12

        vi = h1 + h2 - 2 * mi
        return vi


class VolumeSimilarity(IConfusionMatrixMetric):
    """Represents a volume similarity metric."""

    def __init__(self, metric: str = 'VOLSMTY'):
        """Initializes a new instance of the VolumeSimilarity class.

        Args:
            metric (str): The identification string of the metric.
        """
        super().__init__(metric)

    def calculate(self):
        """Calculates the volume similarity."""

        tp = self.confusion_matrix.tp
        fp = self.confusion_matrix.fp
        fn = self.confusion_matrix.fn

        return 1 - abs(fn - fp) / (2 * tp + fn + fp)



import abc
import typing as t

import numpy as np

import pymia.data.indexexpression as expr


class IndexingStrategy(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def __call__(self, shape) -> t.List[expr.IndexExpression]:
        # return list of indexes by giving shape
        pass

    def __repr__(self) -> str:
        return self.__class__.__name__


class EmptyIndexing(IndexingStrategy):

    def __call__(self, shape) -> t.List[expr.IndexExpression]:
        return [expr.IndexExpression()]


class SliceIndexing(IndexingStrategy):

    def __init__(self, slice_axis: t.Union[int, tuple]=0) -> None:
        if isinstance(slice_axis, int):
            slice_axis = (slice_axis, )
        self.slice_axis = slice_axis

    def __call__(self, shape) -> t.List[expr.IndexExpression]:
        indexing = []
        for axis in self.slice_axis:
            indexing.extend(expr.IndexExpression(i, axis) for i in range(shape[axis]))
        return indexing

    def __repr__(self) -> str:
        return '{} ({})'.format(self.__class__.__name__, self.slice_axis)


class VoxelWiseIndexing(IndexingStrategy):

    def __init__(self, image_dimension: int=3):
        """Initializes a new instance of the VoxelWiseIndexing class.

        Args:
            image_dimension (int): The image dimension without the dimension of the voxels itself.
        """
        self.shape = None
        self.indexing = None
        self.image_dimension = image_dimension

    def __call__(self, shape) -> t.List[expr.IndexExpression]:
        if self.shape == shape:
            return self.indexing

        self.shape = shape  # save for later comparison to avoid calculating indices if the shape is equal
        shape_without_voxel = shape[0:self.image_dimension]
        indices = np.indices(shape_without_voxel)
        indices = indices.reshape((indices.shape[0], np.prod(indices.shape[1:])))
        indices = indices.transpose()
        self.indexing = [expr.IndexExpression(idx.tolist()) for idx in indices]
        return self.indexing


class PatchWiseIndexing(IndexingStrategy):

    def __init__(self, patch_shape: tuple, ignore_incomplete=True) -> None:
        """

        Args:
            patch_shape (tuple):
            ignore_incomplete (bool): If even division of image by patch shape ignore incomplete patch on True.
                Boundary condition.
        """
        super().__init__()
        self.patch_shape = patch_shape
        self.image_dimension = len(patch_shape)
        self.ignore_incomplete = ignore_incomplete
        self.prev_shape = None
        self.prev_indexing = None

    def __call__(self, shape) -> t.List[expr.IndexExpression]:
        if shape == self.prev_shape:
            return self.prev_indexing

        shape_without_voxel = shape[:self.image_dimension]
        index_count = np.divide(shape_without_voxel, self.patch_shape)
        index_count = np.floor(index_count) if self.ignore_incomplete else np.ceil(index_count)
        index_count = index_count.astype('int')

        indices = np.indices(index_count).reshape(index_count.size, -1).T
        index_ranges = np.stack([indices, indices + 1], axis=-1)
        index_ranges *= np.asarray(self.patch_shape)[np.newaxis, :, np.newaxis]
        indexing = [expr.IndexExpression(idx.tolist()) for idx in index_ranges]

        self.prev_indexing = indexing
        self.prev_shape = shape
        return indexing

    def __repr__(self) -> str:
        return '{} (patch shape={}, ignore incomplete={})'.format(self.__class__.__name__,
                                                                  self.patch_shape,
                                                                  self.ignore_incomplete)



import unittest

import numpy as np
import SimpleITK as sitk

import pymia.filtering.postprocessing as fltr


class TestLargestNConnectedComponents(unittest.TestCase):

    def setUp(self):
        # set up image with 3 different connected components
        image = sitk.Image((5, 5), sitk.sitkUInt8)
        image.SetPixel((0, 0), 1)

        image.SetPixel((2, 0), 1)
        image.SetPixel((2, 1), 1)

        image.SetPixel((4, 0), 1)
        image.SetPixel((4, 1), 1)
        image.SetPixel((4, 2), 1)

        self.image = image

    def test_zero_components(self):
        with self.assertRaises(ValueError):
            fltr.LargestNConnectedComponents(0, False)

    def test_one_components(self):
        dut = fltr.LargestNConnectedComponents(1, False)
        result = dut.execute(self.image)

        self.assertEqual(result.GetPixel((4, 0)), 1)
        self.assertEqual(result.GetPixel((4, 1)), 1)
        self.assertEqual(result.GetPixel((4, 2)), 1)

        result_array = sitk.GetArrayFromImage(result)
        self.assertEqual(result_array.sum(), 3)

    def test_two_components(self):
        dut = fltr.LargestNConnectedComponents(2, False)
        result = dut.execute(self.image)

        self.assertEqual(result.GetPixel((2, 0)), 1)
        self.assertEqual(result.GetPixel((2, 1)), 1)
        self.assertEqual(result.GetPixel((4, 0)), 1)
        self.assertEqual(result.GetPixel((4, 1)), 1)
        self.assertEqual(result.GetPixel((4, 2)), 1)

        result_array = sitk.GetArrayFromImage(result)
        self.assertEqual(result_array.sum(), 5)

    def test_three_components(self):
        dut = fltr.LargestNConnectedComponents(3, False)
        result = dut.execute(self.image)

        self.assertEqual(result.GetPixel((0, 0)), 1)
        self.assertEqual(result.GetPixel((2, 0)), 1)
        self.assertEqual(result.GetPixel((2, 1)), 1)
        self.assertEqual(result.GetPixel((4, 0)), 1)
        self.assertEqual(result.GetPixel((4, 1)), 1)
        self.assertEqual(result.GetPixel((4, 2)), 1)

        result_array = sitk.GetArrayFromImage(result)
        self.assertEqual(result_array.sum(), 6)

    def test_four_components(self):
        dut = fltr.LargestNConnectedComponents(3, False)
        result = dut.execute(self.image)

        self.assertEqual(result.GetPixel((0, 0)), 1)
        self.assertEqual(result.GetPixel((2, 0)), 1)
        self.assertEqual(result.GetPixel((2, 1)), 1)
        self.assertEqual(result.GetPixel((4, 0)), 1)
        self.assertEqual(result.GetPixel((4, 1)), 1)
        self.assertEqual(result.GetPixel((4, 2)), 1)

        result_array = sitk.GetArrayFromImage(result)
        self.assertEqual(result_array.sum(), 6)

    def test_consecutive_labels(self):
        dut = fltr.LargestNConnectedComponents(3, True)
        result = dut.execute(self.image)

        self.assertEqual(result.GetPixel((0, 0)), 3)
        self.assertEqual(result.GetPixel((2, 0)), 2)
        self.assertEqual(result.GetPixel((2, 1)), 2)
        self.assertEqual(result.GetPixel((4, 0)), 1)
        self.assertEqual(result.GetPixel((4, 1)), 1)
        self.assertEqual(result.GetPixel((4, 2)), 1)

        result_array = sitk.GetArrayFromImage(result)
        self.assertEqual(result_array.sum(), 10)



"""Provides logging functionality for deep learning algorithms.

Warnings:
    This module is in development and will change with high certainty in the future. Therefore, use carefully!
"""
import abc
import os

import numpy as np
import pymia.deeplearning.visualization as visualization
import tensorboardX as tbx
import tensorflow as tf


class Logger:

    @abc.abstractmethod
    def log_scalar(self, tag: str, value, step: int, is_training: bool = True):
        """Logs a scalar value."""
        pass

    @abc.abstractmethod
    def log_epoch(self, epoch: int, **kwargs):
        pass

    @abc.abstractmethod
    def log_batch(self, step, **kwargs):
        pass

    @abc.abstractmethod
    def log_visualization(self, epoch: int):
        pass

    @abc.abstractmethod
    def close(self):
        pass


class TensorFlowLogger(Logger):

    def __init__(self, log_dir: str, session: tf.Session,
                 epoch_summaries, batch_summaries, visualization_summaries):
        self.session = session

        self.epoch_summaries = epoch_summaries
        self.batch_summaries = batch_summaries
        self.visualization_summaries = visualization_summaries

        # note that the next lines throw a TypeError: Fetch argument None has invalid type <class 'NoneType'>
        # in case no summary has been added
        self.batch_summary_op = tf.summary.merge(self.batch_summaries) if self._is_summary_valid(
            self.batch_summaries) else None
        self.visualization_summary_op1 = tf.summary.merge(self.visualization_summaries[0]) if self._is_summary_valid(
            self.visualization_summaries[0]) else None
        self.visualization_summary_op3 = tf.summary.merge(self.visualization_summaries[3]) if self._is_summary_valid(
            self.visualization_summaries[3]) else None

        # self.summary_op = tf.summary.merge_all()  # note that this should be done AFTER adding operations to tf.summary

        log_dir_train = os.path.join(log_dir, 'train')
        log_dir_valid = os.path.join(log_dir, 'valid')
        os.makedirs(log_dir_train, exist_ok=True)
        os.makedirs(log_dir_valid, exist_ok=True)

        self.writer_train = tf.summary.FileWriter(log_dir_train, session.graph)  # create writer and directly write graph
        self.writer_valid = tf.summary.FileWriter(log_dir_valid)

    def __del__(self):
        self.close()

    def close(self):
        self.writer_train.close()
        self.writer_valid.close()

    @staticmethod
    def _is_summary_valid(summary):
        return not (summary is None or len(summary) == 0)

    def log_scalar(self, tag: str, value, step: int, is_training: bool = True):
        """Logs a scalar value to the TensorBoard.

        Args:
            tag: The scalar's tag.
            value: The value (int, float).
            step: The step the scalar belongs to (global step or epoch).
            is_training: Log using the training writer if True; otherwise, use the validation writer.
        """
        writer = self.writer_train if is_training else self.writer_valid
        summary = tf.Summary(value=[tf.Summary.Value(tag=tag, simple_value=value)])
        writer.add_summary(summary, step)
        writer.flush()

    def log_epoch(self, epoch: int, **kwargs):
        self.log_scalar('loss', kwargs['loss'], epoch, True)
        self.log_scalar('duration', kwargs['duration'], epoch, True)

    def log_batch(self, step, **kwargs):
        if self.batch_summary_op is None:
            return
        # summary_op = tf.summary.merge(self.batch_summaries)
        summary_str = self.session.run(self.batch_summary_op, feed_dict=kwargs['feed_dict'])
        self.writer_train.add_summary(summary_str, step)
        self.writer_train.flush()

    def log_visualization(self, epoch: int):
        if len(self.visualization_summaries[0]) == 0: # todo check if visualization summaries contains any entries
            # todo: wirte Python logger that no visu summary
            return
        # summary_op = tf.summary.merge(self.visualization_summaries[0])
        summary_str = self.session.run(self.visualization_summary_op1)
        self.writer_train.add_summary(summary_str, epoch)

        # todo: clean this up a bit...
        for idx, kernel_name in enumerate(self.visualization_summaries[1]):
            # visualize weights of kernel layer from a middle slice
            kernel_weights = self.session.graph.get_tensor_by_name(kernel_name).eval()
            kernel_weights = np.moveaxis(kernel_weights, -1, 0)  # make last axis the first

            if len(kernel_weights) <= 3:
                grid = visualization.make_grid(kernel_weights)[np.newaxis, :, :, np.newaxis]
                # todo: check case, when does it occur?

            if len(kernel_weights.shape) > 4:
                # todo 3-d convolution, remove last (single) channel
                kernel_weights = kernel_weights[:, :, :, :, 0]
                grid = visualization.make_grid(kernel_weights)[np.newaxis, :, :, np.newaxis]

            if kernel_weights.shape[3] > 1:
                if kernel_weights.shape[1] == 1 and kernel_weights.shape[2] == 1:
                    # 2-D convolution as 1-D convolution, visualize all filters with full channels
                    grid = visualization.make_grid_for_1d_conv(kernel_weights)[np.newaxis, :, :, np.newaxis]
                else:
                    # 2-D convolution with multiple input channels, take example from middle channel
                    slice_num = int(kernel_weights.shape[3] / 2)
                    kernel_weights = kernel_weights[:, :, :, slice_num:slice_num + 1]
                    grid = visualization.make_grid(kernel_weights)[np.newaxis, :, :, np.newaxis]
            else:
                # todo: when??
                grid = visualization.make_grid(kernel_weights)[np.newaxis, :, :, np.newaxis]
            # self.session.run(self.visualization_summaries[2][idx].assign(grid))  # todo(fabianbalsiger): disabled since it adds a new op to the graph, which leads to a growing of the graph...

        # summary_op = tf.summary.merge(self.visualization_summaries[3])
        summary_str = self.session.run(self.visualization_summary_op3)
        self.writer_train.add_summary(summary_str, epoch)
        self.writer_train.flush()

    # Add histograms for trainable variables.
    # for var in tf.trainable_variables():
    #     tf.summary.histogram(var.op.name, var)

    # Add histograms for gradients.
    # for grad, var in grads:
    #     if grad is not None:
    # tf.summary.histogram(var.op.name + '/gradients', grad)


class TorchLogger(Logger):

    def __init__(self, log_dir: str, model,
                 visualize_weights: bool = True, visualize_bias: bool = True, visualize_kernels: bool = True):
        self.model = model

        log_dir_train = os.path.join(log_dir, 'train')
        log_dir_valid = os.path.join(log_dir, 'valid')
        os.makedirs(log_dir_train, exist_ok=True)
        os.makedirs(log_dir_valid, exist_ok=True)

        self.writer_train = tbx.SummaryWriter(log_dir_train)
        self.writer_valid = tbx.SummaryWriter(log_dir_valid)

        self.visualize_weights = visualize_weights
        self.visualize_bias = visualize_bias
        self.visualize_kernels = visualize_kernels

    def __del__(self):
        self.close()

    def close(self):
        self.writer_train.close()
        self.writer_valid.close()

    def log_scalar(self, tag: str, value, step: int, is_training: bool = True):
        writer = self.writer_train if is_training else self.writer_valid
        writer.add_scalar(tag, value, step)

    def log_epoch(self, epoch: int, **kwargs):
        self.log_scalar('loss', kwargs['loss'], epoch, True)
        self.log_scalar('duration', kwargs['duration'], epoch, True)

    def log_batch(self, step, **kwargs):
        pass

    def log_visualization(self, epoch: int):
        for k, v in self.model.state_dict().items():
            # todo(fabianbalsiger): find generic way to identify conv etc. or do we pass a Regex?
            if self.visualize_bias and k.endswith('.bias'):
                # visualize bias of current convolution layer
                self.writer_train.add_histogram(k, v.data.cpu().numpy(), epoch)
            elif k.endswith('.weight'):
                if self.visualize_weights:
                    # visualize weights of current convolution layer
                    self.writer_train.add_histogram(k, v.data.cpu().numpy(), epoch)
                if self.visualize_kernels:

                    # use v.size() to get size of conv and visualize kernel as image...
                    #print(k, v.size()) # eg. 64, 64, 3, 3
                    pass
                    # self.writer.add_image(k, img) # img needs to be of shape (3, H, W) torchvision.utils.make_grid()



"""This module contains classes to set up a filtering pipeline.

All modules in the filter package implement the basic IFilter interface and can be used to set up a pipeline.
"""
import abc
import typing as t

import SimpleITK as sitk


class IFilterParams(metaclass=abc.ABCMeta):
    """Represents a filter parameters interface."""


class IFilter(metaclass=abc.ABCMeta):
    """Filter base class."""

    def __init__(self):
        """Initializes a new instance of the IFilter class."""
        self.verbose = False

    @abc.abstractmethod
    def execute(self, image: sitk.Image, params: IFilterParams=None) -> sitk.Image:
        """Executes a filter on an image.

        Args:
            image (sitk.Image): The image.
            params (IFilterParams): The filter parameters.

        Returns:
            sitk.Image: The filtered image.
        """
        raise NotImplementedError()


class FilterPipeline:
    """Represents a filter pipeline, which sequentially executes filters on images."""

    def __init__(self, filters: t.List[IFilter]=None):
        """Initializes a new instance of the `FilterPipeline` class.

        Args:
            filters (list of IFilter): The filters.
        """
        self.filters = []  # holds the `IFilter`s
        self.params = []  # holds image-specific parameters

        if filters is not None:
            for filter_ in filters:
                self.add_filter(filter_)

    def add_filter(self, filter_: IFilter, params=None):
        """Add a filter to the pipeline.

        Args:
            filter_ (IFilter): A filter.
            params (IFilterParams): The parameters.
        """
        if filter_ is None:
            raise ValueError("The parameter filter needs to be specified.")

        self.filters.append(filter_)
        self.params.append(params)  # params must have the same length as filters

    def set_param(self, params, filter_index: int):
        """Sets an image-specific parameter for a filter.

        Args:
            params (IFilterParams): The parameter(s).
            filter_index (int): The filter's index the parameters belong to.
        """
        self.params[filter_index] = params

    def execute(self, image: sitk.Image) -> sitk.Image:
        """Executes the filter pipeline on an image.

        Args:
            image (sitk.Image): The image.

        Returns:
            sitk.Image: The filtered image.
        """
        for param_index, filter_ in enumerate(self.filters):
            image = filter_.execute(image, self.params[param_index])

        return image

    def __str__(self):
        """Gets a printable string representation.

        Returns:
            str: String representation.
        """

        string = 'FilterPipeline:\n'

        for filter_no, filter_ in enumerate(self.filters):
            string += ' ' + str(filter_no + 1) + '. ' + '    '.join(str(filter_).splitlines(True))

        return string.format(self=self)



import abc
import os

import numpy as np
import h5py

import pymia.data.indexexpression as expr


class Writer(metaclass=abc.ABCMeta):
    """Represents the abstract dataset writer."""

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __del__(self):
        self.close()

    @abc.abstractmethod
    def close(self):
        """Close the writer."""
        pass

    @abc.abstractmethod
    def open(self):
        """Open the writer."""
        pass

    @abc.abstractmethod
    def reserve(self, entry: str, shape: tuple, dtype=None):
        """Reserve space in the dataset for later writing.

        Args:
            entry(str): The dataset entry to be created.
            shape(tuple): The shape to be reserved.
            dtype: The dtype.
        """
        pass

    @abc.abstractmethod
    def fill(self, entry: str, data, index: expr.IndexExpression=None):
        """Fill parts of a reserved dataset entry.

        Args:
            entry(str): The dataset entry to be filled.
            data: The data to write.
            index(expr.IndexExpression): The slicing expression.
        """
        pass

    @abc.abstractmethod
    def write(self, entry: str, data, dtype=None):
        """Create and write entry.

        Args:
            entry(str): The dataset entry to be written.
            data: The data to write.
            dtype: The dtype.
        """
        pass


class Hdf5Writer(Writer):
    """Represents the dataset writer for HDF5 files."""
    str_type = h5py.special_dtype(vlen=str)

    def __init__(self, file_path: str) -> None:
        """Initializes a new instance.

        Args:
            file_path(str): The path to the dataset file to write.
        """
        self.h5 = None  # type: h5py.File
        self.file_path = file_path

    def close(self):
        if self.h5 is not None:
            self.h5.close()
            self.h5 = None

    def open(self):
        self.h5 = h5py.File(self.file_path, libver='latest')

    def reserve(self, entry: str, shape: tuple, dtype=None):
        # special string handling (in order not to use length limited strings)
        if dtype is str or dtype == 'str' or (isinstance(dtype, np.dtype) and dtype.type == np.str_):
            dtype = self.str_type
        self.h5.create_dataset(entry, shape, dtype=dtype)

    def fill(self, entry: str, data, index: expr.IndexExpression=None):
        # special string handling (in order not to use length limited strings)
        if self.h5[entry].dtype is self.str_type:
            data = np.asarray(data, dtype=object)

        if index is None:
            index = expr.IndexExpression()

        self.h5[entry][index.expression] = data

    def write(self, entry: str, data, dtype=None):
        # special string handling (in order not to use length limited strings)
        if dtype is str or dtype == 'str' or (isinstance(dtype, np.dtype) and dtype.type == np.str_):
            dtype = self.str_type
            data = np.asarray(data, dtype=object)
        if entry in self.h5:
            del self.h5[entry]
        self.h5.create_dataset(entry, dtype=dtype, data=data)


def get_writer(file_path: str) -> Writer:
    """ Get the dataset writer corresponding to the file extension.

        Args:
            file_path(str): The path of the dataset file to be written.

        Returns:
            Writer: Writer corresponding to dataset file extension.
        """
    extension = os.path.splitext(file_path)[1]
    if extension not in writer_registry:
        raise ValueError('unknown dataset file extension "{}"'.format(extension))

    return writer_registry[extension](file_path)


writer_registry = {'.h5': Hdf5Writer, '.hdf5': Hdf5Writer}



import unittest
import numpy as np
import SimpleITK as sitk
import pymia.filtering.misc as m


class TestSizeCorrection(unittest.TestCase):

    def test_no_param(self):
        image = sitk.Image(3, 3, sitk.sitkUInt8)
        corrector = m.SizeCorrectionFilter()
        self.assertRaises(ValueError, corrector.execute, image)

    def test_2d(self):
        reference_shape = (4, 6)
        in_shape = (5, 5)
        values = np.random.randint(0, 10, in_shape[::-1])  # since dimesions are switched in itk
        image = sitk.GetImageFromArray(values)

        corrector = m.SizeCorrectionFilter()
        corr_image = corrector.execute(image, params=m.SizeCorrectionParams(reference_shape))

        self.assertEqual(corr_image.GetSize(), reference_shape)
        np_corr_image = sitk.GetArrayFromImage(corr_image)
        self.assertTrue((np_corr_image[:-1] == values[:, :-1]).all())  # dimensions are switched from itk to numpy

    def test_3d(self):
        reference_shape = (4, 4, 3)
        in_shape = (2, 4, 5)
        values = np.random.randint(0, 10, in_shape[::-1])  # since dimesions are switched in itk
        image = sitk.GetImageFromArray(values)

        corrector = m.SizeCorrectionFilter()
        corr_image = corrector.execute(image, params=m.SizeCorrectionParams(reference_shape))

        self.assertEqual(corr_image.GetSize(), reference_shape)
        np_corr_image = sitk.GetArrayFromImage(corr_image)
        self.assertTrue((np_corr_image[:, :, 1:-1] == values[1:-1, :, :]).all())

    def test_crop_and_pad(self):
        reference_shape = (4, 4, 3)
        in_shape = (2, 4, 5)
        values = np.random.randint(0, 10, in_shape[::-1])  # since dimesions are switched in itk
        image = sitk.GetImageFromArray(values)

        corrector = m.SizeCorrectionFilter()
        corr_image = corrector.execute(image, params=m.SizeCorrectionParams(reference_shape))

        self.assertEqual(corr_image.GetSize(), reference_shape)
        np_corr_image = sitk.GetArrayFromImage(corr_image)
        self.assertTrue((np_corr_image[:, :, 1:-1] == values[1:-1, :, :]).all())

    def test_one_sided(self):
        reference_shape = (4, 4, 3)
        in_shape = (2, 4, 5)
        values = np.random.randint(0, 10, in_shape[::-1])  # since dimesions are switched in itk
        image = sitk.GetImageFromArray(values)

        corrector = m.SizeCorrectionFilter(two_sided=False)
        corr_image = corrector.execute(image, params=m.SizeCorrectionParams(reference_shape))

        self.assertEqual(corr_image.GetSize(), reference_shape)
        np_corr_image = sitk.GetArrayFromImage(corr_image)
        self.assertTrue((np_corr_image[:, :, 2:] == values[2:, :, :]).all())

    def test_pad_constant(self):
        reference_shape = (4, 6)
        in_shape = (3, 5)
        values = np.ones(in_shape[::-1])   # since dimesions are switched in itk
        image = sitk.GetImageFromArray(values)

        corrector = m.SizeCorrectionFilter(pad_constant=1)
        corr_image = corrector.execute(image, params=m.SizeCorrectionParams(reference_shape))

        self.assertEqual(corr_image.GetSize(), reference_shape)
        np_corr_image = sitk.GetArrayFromImage(corr_image)
        self.assertTrue((np_corr_image == 1).all())  # dimensions are switched from itk to numpy


class TestRelabel(unittest.TestCase):

    def test_relabel_singles(self):
        image = sitk.Image(4, 4, sitk.sitkUInt8)
        image.SetPixel(1, 1, 5)
        image.SetPixel(1, 2, 5)
        image.SetPixel(3, 3, 1)

        relabel = m.Relabel({3: 5, 2: 1})  # 5 replaced by 3
        re_image = relabel.execute(image)
        np_re_image = sitk.GetArrayFromImage(re_image)
        self.assertFalse((np_re_image == 5).any())
        self.assertEqual(re_image.GetPixel(1, 1), 3)
        self.assertEqual(re_image.GetPixel(1, 2), 3)
        self.assertFalse((np_re_image == 1).any())
        self.assertEqual(re_image.GetPixel(3, 3), 2)

    def test_relabel_multiples(self):
        image = sitk.Image(4, 4, sitk.sitkUInt8)
        image.SetPixel(1, 1, 5)
        image.SetPixel(1, 2, 5)
        image.SetPixel(0, 3, 1)
        image.SetPixel(3, 1, 1)

        relabel = m.Relabel({3: (1, 5)})  # 5 and 1 replaced by 3
        re_image = relabel.execute(image)
        np_re_image = sitk.GetArrayFromImage(re_image)
        self.assertFalse((np_re_image == 5).any())
        self.assertFalse((np_re_image == 1).any())
        self.assertEqual(re_image.GetPixel(1, 1), 3)
        self.assertEqual(re_image.GetPixel(1, 2), 3)
        self.assertEqual(re_image.GetPixel(0, 3), 3)
        self.assertEqual(re_image.GetPixel(3, 1), 3)







"""Enables the plotting of images.

Refer also to `SimpleITK Notebooks
<http://insightsoftwareconsortium.github.io/SimpleITK-Notebooks/10_matplotlib's_imshow.html>`_."""
import os

import matplotlib
matplotlib.use('Agg')
import matplotlib.colors as plt_colors
import matplotlib.pyplot as plt
import numpy as np
import SimpleITK as sitk


def plot_2d_image(path: str, image: np.ndarray, cmap='gray') -> None:
    """Plots a 2-D image.

    Args:
        path (str): The output file path.
        image (np.ndarray): The 2-D image.
        cmap (str): The colormap.
    """

    fig = plt.figure(figsize=image.shape[::-1], dpi=2)  # figure is twice as large as array (in pixels)
    # configure axes such that no boarder is plotted
    # refer to https://github.com/matplotlib/matplotlib/issues/7940/ about how to remove axis from plot
    ax = plt.Axes(fig, [0., 0., 1., 1.])
    ax.set_axis_off()
    ax.margins(0)
    ax.tick_params(which='both', direction='in')

    # plot image
    ax.imshow(image, cmap, interpolation='none')

    fig.add_axes(ax)

    extent = ax.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
    plt.savefig(path, bbox_inches=extent)
    plt.close()


def plot_slice(path: str, image: sitk.Image, slice_no: int) -> None:
    """Plots a slice from a 3-D image.

    Args:
        path (str): The output file path.
        image (sitk.Image): The 3-D image.
        slice_no (int): The slice number.
    """

    slice_ = sitk.GetArrayFromImage(image[:, :, slice_no])
    plot_2d_image(path, slice_)


def plot_2d_segmentation(path: str,
                         image: np.ndarray,
                         ground_truth: np.ndarray,
                         segmentation: np.ndarray,
                         alpha: float=0.5,
                         label: int=1) -> None:
    """Plots a 2-dimensional image with an overlaid mask, which indicates under-, correct-, and over-segmentation.

    Args:
        path (str): The output file path.
        image (np.ndarray): The 2-D image.
        ground_truth (np.ndarray): The 2-D ground truth.
        segmentation (np.ndarray): The 2-D segmentation.
        alpha (float): The alpha blending value, between 0 (transparent) and 1 (opaque).
        label (int): The ground truth and segmentation label.

    Examples:
        >>> img_t2 = np.random.randn(10, 15) * 0.1
        >>> ground_truth = np.zeros((10, 15))
        >>> ground_truth[3:-3, 3:-3] = 1
        >>> segmentation = np.zeros((10, 15))
        >>> segmentation[4:-2, 4:-2] = 1
        >>> plot_2d_segmentation('/your/path/plot_2d_segmentation.png', img_t2, ground_truth, segmentation)
    """

    if not image.shape == ground_truth.shape == segmentation.shape:
        raise ValueError('image, ground_truth, and segmentation must have equal shape')
    if not image.ndim == 2:
        raise ValueError('only 2-dimensional images supported')

    mask = np.zeros(ground_truth.shape)
    mask[np.bitwise_and(ground_truth == label, segmentation != label)] = 1  # under-segmentation
    mask[np.bitwise_and(ground_truth == label, segmentation == label)] = 2  # correct segmentation
    mask[np.bitwise_and(ground_truth != label, segmentation == label)] = 3  # over-segmentation
    masked = np.ma.masked_where(mask == 0, mask)

    fig = plt.figure(figsize=image.shape[::-1], dpi=2)  # figure is twice as large as array (in pixels)
    # configure axes such that no boarder is plotted
    # refer to https://github.com/matplotlib/matplotlib/issues/7940/ about how to remove axis from plot
    ax = plt.Axes(fig, [0., 0., 1., 1.])
    ax.set_axis_off()
    ax.margins(0)
    ax.tick_params(which='both', direction='in')

    # plot image and mask
    ax.imshow(image, 'gray', interpolation='none')
    cm = plt_colors.LinearSegmentedColormap.from_list('rgb',
                                                      [(1, 0, 0), (0, 1, 0), (0, 0, 1)], N=3)  # simple RGB color map
    ax.imshow(masked, interpolation='none', alpha=alpha, cmap=cm)

    fig.add_axes(ax)

    extent = ax.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
    plt.savefig(path, bbox_inches=extent)
    plt.close()


def plot_2d_segmentation_contour(path: str,
                                 image: np.ndarray,
                                 ground_truth: np.ndarray,
                                 segmentation: np.ndarray,
                                 alpha: float=1,
                                 label: int=1) -> None:
    """Plots a 2-dimensional image with overlaid ground truth and segmentation contours.

    The ground truth is plotted in blue and the segmentation in green.
    One can use the SimpleITK `BinaryContourImageFilter
    <https://itk.org/SimpleITKDoxygen/html/classitk_1_1simple_1_1BinaryContourImageFilter.html>`_ to extract a contour.

    Args:
        path (str): The output file path.
        image (np.ndarray): The 2-D image.
        ground_truth (np.ndarray): The 2-D ground truth contour.
        segmentation (np.ndarray): The 2-D segmentation contour.
        alpha (float): The alpha blending value, between 0 (transparent) and 1 (opaque).
        label (int): The ground truth and segmentation label.
    """

    if not image.shape == ground_truth.shape == segmentation.shape:
        raise ValueError('image, ground_truth, and segmentation must have equal shape')
    if not image.ndim == 2:
        raise ValueError('only 2-dimensional images supported')

    mask = np.zeros(ground_truth.shape)
    mask[ground_truth == label] = 1  # set ground truth contour to 1
    mask[segmentation == label] = 2  # set segmentation contour to 2
    masked = np.ma.masked_where(mask == 0, mask)

    fig = plt.figure(figsize=image.shape[::-1], dpi=2)  # figure is twice as large as array (in pixels)
    # configure axes such that no boarder is plotted
    # refer to https://github.com/matplotlib/matplotlib/issues/7940/ about how to remove axis from plot
    ax = plt.Axes(fig, [0., 0., 1., 1.])
    ax.set_axis_off()
    ax.margins(0)
    ax.tick_params(which='both', direction='in')

    # plot image and mask
    ax.imshow(image, 'gray', interpolation='none')
    cm = plt_colors.LinearSegmentedColormap.from_list('rgb',
                                                      [(0.25, 0.5, 1), (0, 1, 0)], N=2)  # simple RGB color map
    ax.imshow(masked, interpolation='none', alpha=alpha, cmap=cm)

    fig.add_axes(ax)

    extent = ax.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
    plt.savefig(path, bbox_inches=extent)
    plt.close()


def plot_2d_segmentation_series(path: str,
                                file_name_suffix: str,
                                image: sitk.Image,
                                ground_truth: sitk.Image,
                                segmentation: sitk.Image,
                                alpha: float=0.5,
                                label: int=1,
                                file_extension: str='.png') -> None:
    """Plots an image with an overlaid mask, which indicates under-, correct-, and over-segmentation.

    Args:
        path (str): The output directory path.
        file_name_suffix (str): The output file name suffix.
        image (sitk.Image): The image.
        ground_truth (sitk.Image): The ground truth.
        segmentation (sitk.Image): The segmentation.
        alpha (float): The alpha blending value, between 0 (transparent) and 1 (opaque).
        label (int): The ground truth and segmentation label.
        file_extension (str): The output file extension (with or without dot).

    Examples:
        >>> img_t2 = sitk.ReadImage('your/path/image.mha')
        >>> ground_truth = sitk.ReadImage('your/path/ground_truth.mha')
        >>> segmentation = sitk.ReadImage('your/path/segmentation.mha')
        >>> plot_2d_segmentation_series('/your/path/', 'mysegmentation', img_t2, ground_truth, segmentation)
    """

    if not image.GetSize() == ground_truth.GetSize() == segmentation.GetSize():
        raise ValueError('image, ground_truth, and segmentation must have equal size')
    if not image.GetDimension() == 3:
        raise ValueError('only 3-dimensional images supported')
    if not image.GetNumberOfComponentsPerPixel() == 1:
        raise ValueError('only scalar images supported')

    img_arr = sitk.GetArrayFromImage(image)
    gt_arr = sitk.GetArrayFromImage(ground_truth)
    seg_arr = sitk.GetArrayFromImage(segmentation)

    os.makedirs(path, exist_ok=True)
    file_extension = file_extension if file_extension.startswith('.') else '.' + file_extension

    for slice_no in range(img_arr.shape[0]):
        full_file_path = os.path.join(path, file_name_suffix + str(slice_no) + file_extension)
        plot_2d_segmentation(full_file_path,
                             img_arr[slice_no, ...],
                             gt_arr[slice_no, ...],
                             seg_arr[slice_no, ...],
                             alpha=alpha,
                             label=label)



"""The evaluator module simplifies the evaluation of segmentation results.

The module provides the possibility of calculate several evaluation metrics in parallel and output them in any format.
"""
import abc
import csv
import logging
import os
from typing import Union

import numpy as np
import SimpleITK as sitk

import pymia.evaluation.metric as pymia_metric


class IEvaluatorWriter(abc.ABC):
    """Represents an evaluator writer interface, which enables to write evaluation results."""

    @abc.abstractmethod
    def write(self, data: list):
        """Writes the evaluation results.

        Args:
            data (list): The evaluation data.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def write_header(self, header: list):
        """Writes the evaluation header.

        Args:
            header (list): The evaluation header.
        """
        raise NotImplementedError


class CSVEvaluatorWriter(IEvaluatorWriter):
    """Represents a CSV (comma-separated values) evaluator writer."""

    def __init__(self, path: str):
        """Initializes a new instance of the CSVEvaluatorWriter class.

        Args:
            path (path): The file path.
        """
        super().__init__()

        self.path = path

        # check file extension
        if not self.path.lower().endswith('.csv'):
            self.path = os.path.join(self.path, '.csv')

        open(self.path, 'w', newline='')  # creates (and overrides an existing) file

    def write(self, data: list):
        """Writes the evaluation results.

        Args:
            data (list): The evaluation data.
        """

        for evaluation in data:
            self.write_line(evaluation)

    def write_header(self, header: list):
        """Writes the evaluation header.

        Args:
            header (list): The evaluation header.
        """

        self.write_line(header)

    def write_line(self, data: list):
        """Writes a line.

        Args:
            data (list): The data.
        """
        with open(self.path, 'a', newline='') as file:
            writer = csv.writer(file, delimiter=';')
            writer.writerow(data)


class ConsoleEvaluatorWriter(IEvaluatorWriter):
    """Represents a console evaluator writer."""

    def __init__(self, precision: int = 3, use_logging: bool = False):
        """Initializes a new instance of the ConsoleEvaluatorWriter class.

        Args:
            precision (int): The float precision.
            use_logging (bool): Indicates whether to use the Python logging module or not.
        """
        super().__init__()

        self.header = None
        self.precision = precision
        self.use_logging = use_logging

    def write(self, data: list):
        """Writes the evaluation results.

        Args:
            data (list of list): The evaluation data,
                e.g. [['PATIENT1', 'BACKGROUND', 0.90], ['PATIENT1', 'TUMOR', '0.62']]
        """

        # format all floats with given precision to str
        out_as_string = [self.header]
        for line in data:
            out_as_string.append([val if isinstance(val, str) else "{0:.{1}f}".format(val, self.precision)
                                  for val in line])

        # determine length of each column for output alignment
        lengths = np.array([list(map(len, row)) for row in out_as_string])
        lengths = lengths.max(0)
        lengths += (len(lengths) - 1) * [2] + [0, ]  # append two spaces except for last column

        # format for output alignment
        out = [['{0:<{1}}'.format(val, lengths[idx]) for idx, val in enumerate(line)] for line in out_as_string]

        if self.use_logging:
            logging.info('\n'.join(''.join(line) for line in out))
        else:
            print('\n'.join(''.join(line) for line in out), sep='', end='\n')

    def write_header(self, header: list):
        """Writes the evaluation header.

        Args:
            header (list of str): The evaluation header.
        """

        self.header = header


class Evaluator:
    """Represents a metric evaluator.

    Examples:

    >>> evaluator = Evaluator(ConsoleEvaluatorWriter(5))
    >>> evaluator.add_writer(CSVEvaluatorWriter('/some/path/to/results.csv'))
    >>> evaluator.add_label(0, 'Background')
    >>> evaluator.add_label(1, 'Structure')
    >>> evaluator.add_metric(DiceCoefficient())
    >>> evaluator.add_metric(VolumeSimilarity())
    >>> evaluator.evaluate(prediction, ground_truth, 'Patient1')
    The console output would be:
              ID       LABEL        DICE     VOLSMTY
        Patient1  Background     0.99955     0.99976
        Patient1       Nerve     0.70692     0.84278
    The results.csv would contain:
    ID;LABEL;DICE;VOLSMTY
    Patient1;Background;0.999548418549;0.999757743496
    Patient1;Nerve;0.70692469107;0.842776093884
    """

    def __init__(self, writer: IEvaluatorWriter=None):
        """Initializes a new instance of the Evaluator class.

        Args:
            writer (IEvaluatorWriter): An evaluator writer.
        """

        self.metrics = []  # list of IMetrics
        self.writers = [writer] if writer is not None else []  # list of IEvaluatorWriters
        self.labels = {}  # dictionary of label: label_str
        self.is_header_written = False

    def add_label(self, label: Union[tuple, int], description: str):
        """Adds a label with its description to the evaluation.

        Args:
            label (Union[tuple, int]): The label or a tuple of labels that should be merged.
            description (str): The label's description.
        """

        self.labels[label] = description

    def add_metric(self, metric: pymia_metric.IMetric):
        """Adds a metric to the evaluation.

        Args:
            metric (pymia_metric.IMetric): The metric.
        """

        self.metrics.append(metric)
        self.is_header_written = False  # header changed due to new metric

    def add_writer(self, writer: IEvaluatorWriter):
        """Adds a writer to the evaluation.

        Args:
            writer (IEvaluatorWriter): The writer.
        """

        self.writers.append(writer)
        self.is_header_written = False  # re-write header

    def evaluate(self, image: Union[sitk.Image, np.ndarray], ground_truth: Union[sitk.Image, np.ndarray],
                 evaluation_id: str):
        """Evaluates the metrics on the provided image and ground truth image.

        Args:
            image (sitk.Image): The segmented image.
            ground_truth (sitk.Image): The ground truth image.
            evaluation_id (str): The identification of the evaluation.

        Raises:
            ValueError: If no labels are defined (see add_label).
        """

        if not self.labels:
            raise ValueError('No labels to evaluate defined')

        if isinstance(image, sitk.Image) and image.GetNumberOfComponentsPerPixel() > 1:
            raise ValueError('Image has more than one component per pixel')

        if isinstance(ground_truth, sitk.Image) and ground_truth.GetNumberOfComponentsPerPixel() > 1:
            raise ValueError('Image has more than one component per pixel')

        if not self.is_header_written:
            self.write_header()

        image_array = sitk.GetArrayFromImage(image) if isinstance(image, sitk.Image) else image
        ground_truth_array = sitk.GetArrayFromImage(ground_truth) if isinstance(ground_truth, sitk.Image) else ground_truth

        results = []  # clear results

        for label, label_str in self.labels.items():
            label_results = [evaluation_id, label_str]

            # get only current label
            predictions = np.in1d(image_array.ravel(), label, True).reshape(image_array.shape).astype(np.uint8)
            labels = np.in1d(ground_truth_array.ravel(), label, True).reshape(ground_truth_array.shape).astype(np.uint8)

            # calculate the confusion matrix for IConfusionMatrixMetric
            confusion_matrix = pymia_metric.ConfusionMatrix(predictions, labels)

            # flag indicating whether the images have been converted for ISimpleITKImageMetric
            converted_to_image = False
            predictions_as_image = None
            labels_as_image = None

            # for distance metric
            distances = None

            # calculate the metrics
            for param_index, metric in enumerate(self.metrics):
                if isinstance(metric, pymia_metric.IConfusionMatrixMetric):
                    metric.confusion_matrix = confusion_matrix
                elif isinstance(metric, pymia_metric.INumpyArrayMetric):
                    metric.ground_truth = labels
                    metric.segmentation = predictions
                elif isinstance(metric, pymia_metric.ISimpleITKImageMetric):
                    if not converted_to_image:
                        if not isinstance(image, sitk.Image):
                            raise ValueError('SimpleITK image is required for SimpleITK-based metrics')
                        predictions_as_image = sitk.GetImageFromArray(predictions)
                        predictions_as_image.CopyInformation(image)
                        labels_as_image = sitk.GetImageFromArray(labels)
                        labels_as_image.CopyInformation(ground_truth)
                        converted_to_image = True

                    metric.ground_truth = labels_as_image
                    metric.segmentation = predictions_as_image
                elif isinstance(metric, pymia_metric.IDistanceMetric):
                    if distances is None:
                        if isinstance(image, sitk.Image):
                            spacing = image.GetSpacing()[::-1]
                        else:
                            spacing = (1.0,) * labels.ndim  # use isotropic spacing of 1 mm

                        distances = pymia_metric.Distances(predictions, labels, spacing)

                    metric.distances = distances

                label_results.append(metric.calculate())

            results.append(label_results)

        # write the results
        for writer in self.writers:
            writer.write(results)

    def write_header(self):
        """Writes the header."""

        header = ['ID', 'LABEL']

        for metric in self.metrics:
            header.append(str(metric))

        for writer in self.writers:
            writer.write_header(header)

        self.is_header_written = True



import unittest

import numpy as np
import SimpleITK as sitk

import pymia.filtering.filter as pymia_fltr
import pymia.filtering.preprocessing as pymia_fltr_prep


class TestNormalizeZScore(unittest.TestCase):

    def setUp(self):
        # set up image
        image = sitk.Image((4, 1), sitk.sitkUInt8)
        image.SetPixel((0, 0), 1)
        image.SetPixel((1, 0), 2)
        image.SetPixel((2, 0), 3)
        image.SetPixel((3, 0), 4)

        self.image = image

        # test_case = [1, 2, 3, 4]
        # not in R, so tested by using:
        #    (test_case[i] - mean(test_case, axis=0)) / sqrt(var(test_case) * 3/4)
        self.desired = np.array([[-1.3416407864999, -0.44721359549996, 0.44721359549996, 1.3416407864999]], np.float64)

    def test_normalization(self):
        dut = pymia_fltr_prep.NormalizeZScore()
        out = dut.execute(self.image)
        out_arr = sitk.GetArrayFromImage(out)

        np.testing.assert_array_almost_equal(self.desired, out_arr, decimal=12)

    def test_normalization_with_param(self):
        dut = pymia_fltr_prep.NormalizeZScore()
        out = dut.execute(self.image, pymia_fltr.IFilterParams())
        out_arr = sitk.GetArrayFromImage(out)

        np.testing.assert_array_almost_equal(self.desired, out_arr, decimal=12)

    def test_image_properties(self):
        dut = pymia_fltr_prep.NormalizeZScore()
        out = dut.execute(self.image)
        self.assertEqual(self.image.GetSize(), out.GetSize())
        self.assertEqual(self.image.GetOrigin(), out.GetOrigin())
        self.assertEqual(self.image.GetSpacing(), out.GetSpacing())
        self.assertEqual(self.image.GetDirection(), out.GetDirection())
        self.assertEqual(self.image.GetDimension(), out.GetDimension())
        self.assertEqual(self.image.GetNumberOfComponentsPerPixel(), out.GetNumberOfComponentsPerPixel())

        self.assertEqual(sitk.sitkFloat64, out.GetPixelID())



import os
import typing

import numpy as np

import pymia.data.indexexpression as expr
import pymia.data.conversion as conv
import pymia.data.subjectfile as subj
import pymia.data.definition as df
from . import writer as wr


class Callback:

    def on_start(self, params: dict):
        pass

    def on_end(self, params: dict):
        pass

    def on_subject(self, params: dict):
        pass


class ComposeCallback(Callback):

    def __init__(self, callbacks: typing.List[Callback]) -> None:
        self.callbacks = callbacks

    def on_start(self, params: dict):
        for c in self.callbacks:
            c.on_start(params)

    def on_end(self, params: dict):
        for c in self.callbacks:
            c.on_end(params)

    def on_subject(self, params: dict):
        for c in self.callbacks:
            c.on_subject(params)


class MonitoringCallback(Callback):

    def on_start(self, params: dict):
        print('start dataset creation')

    def on_subject(self, params: dict):
        index = params['subject_index']
        subject_files = params['subject_files']
        print('[{}/{}] {}'.format(index + 1, len(subject_files), subject_files[index].subject))

    def on_end(self, params: dict):
        print('dataset creation finished')


class WriteDataCallback(Callback):

    def __init__(self, writer: wr.Writer) -> None:
        self.writer = writer

    def on_subject(self, params: dict):
        subject_files = params['subject_files']
        subject_index = params['subject_index']

        max_digits = len(str(len(subject_files)))
        index_str = '{{:0{}}}'.format(max_digits).format(subject_index)

        for category in params['categories']:
            data = params[category]
            self.writer.write('{}/{}'.format(df.DATA_PLACEHOLDER.format(category), index_str), data, dtype=data.dtype)


class WriteSubjectCallback(Callback):

    def __init__(self, writer: wr.Writer) -> None:
        self.writer = writer

    def on_start(self, params: dict):
        subject_count = len(params['subject_files'])
        self.writer.reserve(df.SUBJECT, (subject_count,), str)

    def on_subject(self, params: dict):
        subject_files = params['subject_files']
        subject_index = params['subject_index']

        subject = subject_files[subject_index].subject
        self.writer.fill(df.SUBJECT, subject, expr.IndexExpression(subject_index))


class WriteImageInformationCallback(Callback):

    def __init__(self, writer: wr.Writer, category='images') -> None:
        self.writer = writer
        self.category = category
        self.new_subject = False

    def on_start(self, params: dict):
        subject_count = len(params['subject_files'])
        self.writer.reserve(df.INFO_SHAPE, (subject_count, 3), dtype=np.uint16)
        self.writer.reserve(df.INFO_ORIGIN, (subject_count, 3), dtype=np.float)
        self.writer.reserve(df.INFO_DIRECTION, (subject_count, 9), dtype=np.float)
        self.writer.reserve(df.INFO_SPACING, (subject_count, 3), dtype=np.float)

    def on_subject(self, params: dict):
        subject_index = params['subject_index']
        properties = params['{}_properties'.format(self.category)]  # type: conv.ImageProperties

        self.writer.fill(df.INFO_SHAPE, properties.size, expr.IndexExpression(subject_index))
        self.writer.fill(df.INFO_ORIGIN, properties.origin, expr.IndexExpression(subject_index))
        self.writer.fill(df.INFO_DIRECTION, properties.direction, expr.IndexExpression(subject_index))
        self.writer.fill(df.INFO_SPACING, properties.spacing, expr.IndexExpression(subject_index))


class WriteNamesCallback(Callback):

    def __init__(self, writer: wr.Writer) -> None:
        self.writer = writer

    def on_start(self, params: dict):
        for category in params['categories']:
            self.writer.write(df.NAMES_PLACEHOLDER.format(category), params['{}_names'.format(category)], dtype='str')


class WriteFilesCallback(Callback):

    def __init__(self, writer: wr.Writer) -> None:
        self.writer = writer
        self.file_root = None

    @staticmethod
    def _get_common_path(subject_files):
        def get_subject_common(subject_file: subj.SubjectFile):
            return os.path.commonpath(list(subject_file.get_all_files().values()))
        return os.path.commonpath([get_subject_common(sf) for sf in subject_files])

    def on_start(self, params: dict):
        subject_files = params['subject_files']
        self.file_root = self._get_common_path(subject_files)
        self.writer.write(df.FILES_ROOT, self.file_root, dtype='str')

        for category in params['categories']:
            self.writer.reserve(df.FILES_PLACEHOLDER.format(category),
                                (len(subject_files), len(params['{}_names'.format(category)])), dtype='str')

    def on_subject(self, params: dict):
        subject_index = params['subject_index']
        subject_files = params['subject_files']

        subject_file = subject_files[subject_index]  # type: subj.SubjectFile

        for category in params['categories']:
            for index, file_name in enumerate(subject_file.categories[category].entries.values()):
                relative_path = os.path.relpath(file_name, self.file_root)
                index_expr = expr.IndexExpression(indexing=[subject_index, index], axis=(0, 1))
                self.writer.fill(df.FILES_PLACEHOLDER.format(category), relative_path, index_expr)


def get_default_callbacks(writer: wr.Writer) -> ComposeCallback:
    return ComposeCallback([MonitoringCallback(),
                            WriteDataCallback(writer),
                            WriteFilesCallback(writer),
                            WriteNamesCallback(writer),
                            WriteImageInformationCallback(writer),
                            WriteSubjectCallback(writer)])



from .reader import (Reader, Hdf5Reader, get_reader)
from .indexing import (IndexingStrategy, SliceIndexing, VoxelWiseIndexing, EmptyIndexing, PatchWiseIndexing)
from .dataset import ParameterizableDataset
from .extractor import (Extractor, DataExtractor, FilesExtractor, NamesExtractor, SubjectExtractor, IndexingExtractor,
                        SelectiveDataExtractor, RandomDataExtractor, ComposeExtractor,
                        ImagePropertiesExtractor, PadDataExtractor, ImageShapeExtractor)
from .sample import (select_indices, SubsetSequentialSampler, NonBlackSelection, SelectionStrategy, ComposeSelection,
                     SubjectSelection, WithForegroundSelection, PercentileSelection, NonConstantSelection)

# pytorch class forwarding
from torch.utils.data.sampler import (WeightedRandomSampler,SequentialSampler,Sampler, RandomSampler, BatchSampler,
                                      SubsetRandomSampler)
from torch.utils.data import (Dataset, ConcatDataset, TensorDataset, DataLoader)




"""Enables the enhancement of images before their use with other algorithms."""
import SimpleITK as sitk

import pymia.filtering.filter as pymia_fltr


class BiasFieldCorrectorParams(pymia_fltr.IFilterParams):
    """Bias field correction filter parameters."""

    def __init__(self, mask: sitk.Image):
        """Initializes a new instance of the BiasFieldCorrectorParams class.

        Args:
            mask (sitk.Image): A mask image (0=background; 1=mask).

        Examples:
        To generate a default mask use Otsu's thresholding:

        >>> sitk.OtsuThreshold(image, 0, 1, 200)
        """
        self.mask = mask


class BiasFieldCorrector(pymia_fltr.IFilter):
    """Represents a bias field correction filter."""

    def __init__(self, shrink_factor=1, convergence_threshold=0.001, max_iterations=(50, 50, 50, 50),
                 fullwidth_at_halfmax=0.15, fiter_noise=0.01, histogram_bins=200, control_points=(4, 4, 4),
                 spline_order=3):
        """Initializes a new instance of the BiasFieldCorrector class.

        Args:
            shrink_factor (float): The shrink factor. A higher factor decreases the computational time.
            convergence_threshold (float): The threshold to stop the optimizer.
            max_iterations (list of int): The maximum number of optimizer iterations at each level.
            fullwidth_at_halfmax (float): ?
            fiter_noise (float): Wiener filter noise.
            histogram_bins (int): Number of histogram bins.
            control_points (list of int): The number of spline control points.
            spline_order (int): The spline order.
        """
        super().__init__()
        self.shrink_factor = shrink_factor
        self.convergence_threshold = convergence_threshold
        self.max_iterations = max_iterations
        self.fullwidth_at_halfmax = fullwidth_at_halfmax
        self.filter_noise = fiter_noise
        self.histogram_bins = histogram_bins
        self.control_points = control_points
        self.spline_order = spline_order

    def execute(self, image: sitk.Image, params: BiasFieldCorrectorParams=None) -> sitk.Image:
        """Executes a bias field correction on an image.

        Args:
            image (sitk.Image): The image.
            params (BiasFieldCorrectorParams): The bias field correction filter parameters.

        Returns:
            sitk.Image: The bias field corrected image.
        """

        mask = params.mask if params is not None else sitk.OtsuThreshold(image, 0, 1, 200)
        if self.shrink_factor > 1:
            raise ValueError('shrinking is not supported yet')
        return sitk.N4BiasFieldCorrection(image, mask, self.convergence_threshold, self.max_iterations,
                                          self.fullwidth_at_halfmax, self.filter_noise, self.histogram_bins,
                                          self.control_points, self.spline_order)

    def __str__(self):
        """Gets a printable string representation.

        Returns:
            str: String representation.
        """
        return 'BiasFieldCorrector:\n' \
               ' shrink_factor:         {self.shrink_factor}\n' \
               ' convergence_threshold: {self.convergence_threshold}\n' \
               ' max_iterations:        {self.max_iterations}\n' \
               ' fullwidth_at_halfmax:  {self.fullwidth_at_halfmax}\n' \
               ' filter_noise:          {self.filter_noise}\n' \
               ' histogram_bins:        {self.histogram_bins}\n' \
               ' control_points:        {self.control_points}\n' \
               ' spline_order:          {self.spline_order}\n' \
            .format(self=self)


class GradientAnisotropicDiffusion(pymia_fltr.IFilter):
    """Represents a gradient anisotropic diffusion filter."""

    def __init__(self,
                 time_step: float=0.125,
                 conductance: int=3,
                 conductance_scaling_update_interval: int=1,
                 no_iterations: int=5):
        """Initializes a new instance of the GradientAnisotropicDiffusion class.

        Args:
            time_step (float): The time step.
            conductance (int): The conductance (the higher the smoother the edges).
            conductance_scaling_update_interval: TODO
            no_iterations (int): Number of iterations.
        """
        super().__init__()
        self.time_step = time_step
        self.conductance = conductance
        self.conductance_scaling_update_interval = conductance_scaling_update_interval
        self.no_iterations = no_iterations

    def execute(self, image: sitk.Image, params: pymia_fltr.IFilterParams=None) -> sitk.Image:
        """Executes a gradient anisotropic diffusion on an image.

        Args:
            image (sitk.Image): The image.
            params (IFilterParams): The parameters (unused).

        Returns:
            sitk.Image: The smoothed image.
        """
        return sitk.GradientAnisotropicDiffusion(sitk.Cast(image, sitk.sitkFloat32),
                                                 self.time_step,
                                                 self.conductance,
                                                 self.conductance_scaling_update_interval,
                                                 self.no_iterations)

    def __str__(self):
        """Gets a printable string representation.

        Returns:
            str: String representation.
        """
        return 'GradientAnisotropicDiffusion:\n' \
               ' time_step:                           {self.time_step}\n' \
               ' conductance:                         {self.conductance}\n' \
               ' conductance_scaling_update_interval: {self.conductance_scaling_update_interval}\n' \
               ' no_iterations:                       {self.no_iterations}\n' \
            .format(self=self)


class NormalizeZScore(pymia_fltr.IFilter):
    """Represents a z-score normalization filter."""

    def __init__(self):
        """Initializes a new instance of the NormalizeZScore class."""
        super().__init__()

    def execute(self, image: sitk.Image, params: pymia_fltr.IFilterParams=None) -> sitk.Image:
        """Executes a z-score normalization on an image.

        Args:
            image (sitk.Image): The image.
            params (IFilterParams): The parameters (unused).

        Returns:
            sitk.Image: The normalized image.
        """

        img_arr = sitk.GetArrayFromImage(image)

        mean = img_arr.mean()
        std = img_arr.std()
        img_arr = (img_arr - mean) / std

        img_out = sitk.GetImageFromArray(img_arr)
        img_out.CopyInformation(image)

        return img_out

    def __str__(self):
        """Gets a printable string representation.

        Returns:
            str: String representation.
        """
        return 'NormalizeZScore:\n' \
            .format(self=self)


class RescaleIntensity(pymia_fltr.IFilter):
    """Represents a rescale intensity filter."""

    def __init__(self, min_intensity, max_intensity):
        """Initializes a new instance of the RescaleIntensity class.

        Args:
            min_intensity (int): The min intensity value.
            max_intensity (int): The max intensity value.
        """
        super().__init__()
        self.min_intensity = min_intensity
        self.max_intensity = max_intensity

    def execute(self, image: sitk.Image, params: pymia_fltr.IFilterParams=None) -> sitk.Image:
        """Executes an intensity rescaling on an image.

        Args:
            image (sitk.Image): The image.
            params (IFilterParams): The parameters (unused).

        Returns:
            sitk.Image: The intensity rescaled image.
        """

        return sitk.RescaleIntensity(image, self.min_intensity, self.max_intensity)

    def __str__(self):
        """Gets a printable string representation.

        Returns:
            str: String representation.
        """
        return 'RescaleIntensity:\n' \
               ' min_intensity: {self.min_intensity}\n' \
               ' max_intensity: {self.max_intensity}\n' \
            .format(self=self)


class HistogramMatcherParams(pymia_fltr.IFilterParams):
    """Histogram matching filter parameters."""

    def __init__(self, reference_image: sitk.Image):
        """Initializes a new instance of the HistogramMatcherParams class.

        Args:
            reference_image (sitk.Image): Reference image for the matching.
        """
        self.reference_image = reference_image


class HistogramMatcher(pymia_fltr.IFilter):
    """A method to align the intensity ranges of images."""

    def __init__(self, histogram_levels=256, match_points=1, threshold_mean_intensity=True):
        """Initializes a new instance of the HistogramMatcher class.

        Args:
            histogram_levels (int): Number of histogram levels.
            match_points (int): Number of match points.
            threshold_mean_intensity (bool): Threshold at mean intensity.
        """
        super().__init__()
        self.histogram_levels = histogram_levels
        self.match_points = match_points
        self.threshold_mean_intensity = threshold_mean_intensity

    def execute(self, image: sitk.Image, params: HistogramMatcherParams = None) -> sitk.Image:
        """Matches the image intensity histogram to a reference.

        Args:
            image (sitk.Image): The image.
            params (HistogramMatcherParams): The parameters.

        Returns:
            sitk.Image: The filtered image.
        """
        if params is None:
            raise ValueError('Parameter with reference image is required')

        return sitk.HistogramMatching(image, params.reference_image, self.histogram_levels, self.match_points,
                                      self.threshold_mean_intensity)

    def __str__(self):
        """Gets a printable string representation.

        Returns:
            str: String representation.
        """
        return 'HistogramMatcher:\n' \
               ' histogram_levels:         {self.histogram_levels}\n' \
               ' match_points:             {self.match_points}\n' \
               ' threshold_mean_intensity: {self.threshold_mean_intensity}\n' \
            .format(self=self)



"""The registration module contains classes for image registration.

Image registration aims to align two images using a particular transformation.
pymia currently supports multi-modal rigid registration, i.e. align two images of different modalities
using a rigid transformation (rotation, translation, reflection, or their combination).

See Also:
    `ITK Registration <https://itk.org/Doxygen/html/RegistrationPage.html>`_
    `ITK Software Guide Registration <https://itk.org/ITKSoftwareGuide/html/Book2/ITKSoftwareGuide-Book2ch3.html>`_
"""
import abc
import enum
import os
import typing as t

import matplotlib
matplotlib.use('Agg')  # use matplotlib without having a window appear
import matplotlib.pyplot as plt
import numpy as np
import SimpleITK as sitk

import pymia.filtering.filter as pymia_fltr


class RegistrationType(enum.Enum):
    """Represents the registration transformation type."""
    AFFINE = 1
    SIMILARITY = 2
    RIGID = 3
    BSPLINE = 4


class RegistrationCallback(metaclass=abc.ABCMeta):
    """Represents the abstract handler for the registration callbacks."""

    def __init__(self) -> None:
        """Initializes a new instance of the abstract RegistrationCallback class."""
        self.registration_method = None
        self.fixed_image = None
        self.moving_image = None
        self.transform = None

    def set_params(self, registration_method: sitk.ImageRegistrationMethod,
                   fixed_image: sitk.Image,
                   moving_image: sitk.Image,
                   transform: sitk.Transform):
        """Sets the parameters that might be used during the callbacks

        Args:
            registration_method (sitk.ImageRegistrationMethod): The registration method.
            fixed_image (sitk.Image): The fixed image.
            moving_image (sitk.Image): The moving image.
            transform (sitk.Transform): The transformation.
        """
        self.registration_method = registration_method
        self.fixed_image = fixed_image
        self.moving_image = moving_image
        self.transform = transform
        # link the callback functions to the events
        self.registration_method.AddCommand(sitk.sitkStartEvent, self.registration_started)
        self.registration_method.AddCommand(sitk.sitkEndEvent, self.registration_ended)
        self.registration_method.AddCommand(sitk.sitkMultiResolutionIterationEvent,
                                            self.registration_resolution_changed)
        self.registration_method.AddCommand(sitk.sitkIterationEvent, self.registration_iteration_ended)

    def registration_ended(self):
        """Callback for the EndEvent."""
        pass

    def registration_started(self):
        """Callback for the StartEvent."""
        pass

    def registration_resolution_changed(self):
        """Callback for the MultiResolutionIterationEvent."""
        pass

    def registration_iteration_ended(self):
        """Callback for the IterationEvent."""
        pass


class MultiModalRegistrationParams(pymia_fltr.IFilterParams):
    """Represents parameters for the multi-modal rigid registration."""

    def __init__(self, fixed_image: sitk.Image, fixed_image_mask: sitk.Image=None,
                 callbacks: t.List[RegistrationCallback]=None):
        """Initializes a new instance of the MultiModalRegistrationParams class.

        Args:
            fixed_image (sitk.Image): The fixed image for the registration.
            fixed_image_mask (sitk.Image): A mask for the fixed image to limit the registration.
            callbacks (t.List[RegistrationCallback]): Path to the directory where to plot the registration
                progress if any. Note that this increases the computational time.
        """

        self.fixed_image = fixed_image
        self.fixed_image_mask = fixed_image_mask
        self.callbacks = callbacks


class MultiModalRegistration(pymia_fltr.IFilter):
    """Represents a multi-modal image registration filter.

    The filter estimates a 3-dimensional rigid or affine transformation between images of different modalities using
    - Mutual information similarity metric
    - Linear interpolation
    - Gradient descent optimization

    Examples:

    The following example shows the usage of the MultiModalRegistration class.

    >>> fixed_image = sitk.ReadImage('/path/to/image/fixed.mha')
    >>> moving_image = sitk.ReadImage('/path/to/image/moving.mha')
    >>> registration = MultiModalRegistration()  # specify parameters to your needs
    >>> parameters = MultiModalRegistrationParams(fixed_image)
    >>> registered_image = registration.execute(moving_image, parameters)
    """

    def __init__(self,
                 registration_type: RegistrationType=RegistrationType.RIGID,
                 number_of_histogram_bins: int=200,
                 learning_rate: float=1.0,
                 step_size: float=0.001,
                 number_of_iterations: int=200,
                 relaxation_factor: float=0.5,
                 shrink_factors: [int]=(2, 1, 1),
                 smoothing_sigmas: [float]=(2, 1, 0),
                 sampling_percentage: float=0.2,
                 sampling_seed: int=sitk.sitkWallClock,
                 resampling_interpolator=sitk.sitkBSpline):
        """Initializes a new instance of the MultiModalRegistration class.

        Args:
            registration_type (RegistrationType): The type of the registration ('rigid' or 'affine').
            number_of_histogram_bins (int): The number of histogram bins.
            learning_rate (float): The optimizer's learning rate.
            step_size (float): The optimizer's step size. Each step in the optimizer is at least this large.
            number_of_iterations (int): The maximum number of optimization iterations.
            relaxation_factor (float): The relaxation factor to penalize abrupt changes during optimization.
            shrink_factors ([int]): The shrink factors at each shrinking level (from high to low).
            smoothing_sigmas ([int]):  The Gaussian sigmas for smoothing at each shrinking level (in physical units).
            sampling_percentage (float): Fraction of voxel of the fixed image that will be used for registration (0, 1].
                Typical values range from 0.01 (1 %) for low detail images to 0.2 (20 %) for high detail images.
                The higher the fraction, the higher the computational time.
            sampling_seed: The seed for reproducible behavior.
            resampling_interpolator: Interpolation to be applied while resampling the image by the determined
                transformation.
        """
        super().__init__()

        if len(shrink_factors) != len(smoothing_sigmas):
            raise ValueError("shrink_factors and smoothing_sigmas need to be same length")

        self.registration_type = registration_type
        self.number_of_histogram_bins = number_of_histogram_bins
        self.learning_rate = learning_rate
        self.step_size = step_size
        self.number_of_iterations = number_of_iterations
        self.relaxation_factor = relaxation_factor
        self.shrink_factors = shrink_factors
        self.smoothing_sigmas = smoothing_sigmas
        self.sampling_percentage = sampling_percentage
        self.sampling_seed = sampling_seed
        self.resampling_interpolator = resampling_interpolator

        registration = sitk.ImageRegistrationMethod()

        # similarity metric
        # will compare how well the two images match each other
        # registration.SetMetricAsJointHistogramMutualInformation(self.number_of_histogram_bins, 1.5)
        registration.SetMetricAsMattesMutualInformation(self.number_of_histogram_bins)
        registration.SetMetricSamplingStrategy(registration.RANDOM)
        registration.SetMetricSamplingPercentage(self.sampling_percentage, self.sampling_seed)

        # An image gradient calculator based on ImageFunction is used instead of image gradient filters
        # set to True uses GradientRecursiveGaussianImageFilter
        # set to False uses CentralDifferenceImageFunction
        # see also https://itk.org/Doxygen/html/classitk_1_1ImageToImageMetricv4.html
        registration.SetMetricUseFixedImageGradientFilter(False)
        registration.SetMetricUseMovingImageGradientFilter(False)

        # interpolator
        # will evaluate the intensities of the moving image at non-rigid positions
        registration.SetInterpolator(sitk.sitkLinear)

        # optimizer
        # is required to explore the parameter space of the transform in search of optimal values of the metric
        if self.registration_type == RegistrationType.BSPLINE:
            registration.SetOptimizerAsLBFGSB()
        else:
            registration.SetOptimizerAsRegularStepGradientDescent(learningRate=self.learning_rate,
                                                                  minStep=self.step_size,
                                                                  numberOfIterations=self.number_of_iterations,
                                                                  relaxationFactor=self.relaxation_factor,
                                                                  gradientMagnitudeTolerance=1e-4,
                                                                  estimateLearningRate=registration.EachIteration,
                                                                  maximumStepSizeInPhysicalUnits=0.0)
        registration.SetOptimizerScalesFromPhysicalShift()

        # setup for the multi-resolution framework
        registration.SetShrinkFactorsPerLevel(self.shrink_factors)
        registration.SetSmoothingSigmasPerLevel(self.smoothing_sigmas)
        registration.SmoothingSigmasAreSpecifiedInPhysicalUnitsOn()

        self.registration = registration
        self.transform = None

    def execute(self, image: sitk.Image, params: MultiModalRegistrationParams=None) -> sitk.Image:
        """Executes a multi-modal rigid registration.

        Args:
            image (sitk.Image): The moving image.
            params (MultiModalRegistrationParams): The parameters, which contain the fixed image.

        Returns:
            sitk.Image: The registered image.
        """

        if params is None:
            raise ValueError("params is not defined")
        dimension = image.GetDimension()
        if dimension not in (2, 3):
            raise ValueError('Image dimension {} is not among the accepted (2, 3)'.format(dimension))

        # set a transform that is applied to the moving image to initialize the registration
        if self.registration_type == RegistrationType.BSPLINE:
            transform_domain_mesh_size = [10] * image.GetDimension()
            initial_transform = sitk.BSplineTransformInitializer(params.fixed_image, transform_domain_mesh_size)
        else:
            if self.registration_type == RegistrationType.RIGID:
                transform_type = sitk.VersorRigid3DTransform() if dimension == 3 else sitk.Euler2DTransform()
            elif self.registration_type == RegistrationType.AFFINE:
                transform_type = sitk.AffineTransform(dimension)
            elif self.registration_type == RegistrationType.SIMILARITY:
                transform_type = sitk.Similarity3DTransform() if dimension == 3 else sitk.Similarity2DTransform()
            else:
                raise ValueError('not supported registration_type')

            initial_transform = sitk.CenteredTransformInitializer(sitk.Cast(params.fixed_image,
                                                                            image.GetPixelIDValue()),
                                                                  image,
                                                                  transform_type,
                                                                  sitk.CenteredTransformInitializerFilter.GEOMETRY)

        self.registration.SetInitialTransform(initial_transform, inPlace=True)

        if params.fixed_image_mask:
            self.registration.SetMetricFixedMask(params.fixed_image_mask)

        if params.callbacks is not None:
            for callback in params.callbacks:
                callback.set_params(self.registration, params.fixed_image, image, initial_transform)

        self.transform = self.registration.Execute(sitk.Cast(params.fixed_image, sitk.sitkFloat32),
                                                   sitk.Cast(image, sitk.sitkFloat32))

        if self.verbose:
            print('MultiModalRegistration:\n Final metric value: {0}'.format(self.registration.GetMetricValue()))
            print(' Optimizer\'s stopping condition, {0}'.format(
                self.registration.GetOptimizerStopConditionDescription()))
        elif self.number_of_iterations == self.registration.GetOptimizerIteration():
            print('MultiModalRegistration: Optimizer terminated at number of iterations and did not converge!')

        return sitk.Resample(image, params.fixed_image, self.transform, self.resampling_interpolator, 0.0,
                             image.GetPixelIDValue())

    def __str__(self):
        """Gets a nicely printable string representation.

        Returns:
            str: The string representation.
        """

        return 'MultiModalRegistration:\n' \
               ' registration_type:        {self.registration_type}\n' \
               ' number_of_histogram_bins: {self.number_of_histogram_bins}\n' \
               ' learning_rate:            {self.learning_rate}\n' \
               ' step_size:                {self.step_size}\n' \
               ' number_of_iterations:     {self.number_of_iterations}\n' \
               ' relaxation_factor:        {self.relaxation_factor}\n' \
               ' shrink_factors:           {self.shrink_factors}\n' \
               ' smoothing_sigmas:         {self.smoothing_sigmas}\n' \
               ' sampling_percentage:      {self.sampling_percentage}\n' \
               ' resampling_interpolator:  {self.resampling_interpolator}\n' \
            .format(self=self)


class PlotCallback(RegistrationCallback):
    """Represents a plotter for SimpleITK registrations."""

    def __init__(self, plot_dir: str, file_name_prefix: str='', slice_no: int=-1) -> None:
        """
        Args:
            plot_dir (str): Path to the directory where to save the plots.
            file_name_prefix (str): The file name prefix for the plots.
            slice_no (int): The slice number to plot (affects only 3-D images). -1 means to use the middle slice.
        """
        super().__init__()
        self.plot_dir = plot_dir
        self.file_name_prefix = file_name_prefix
        self.slice_no = slice_no

        self.metric_values = []
        self.resolution_iterations = []

    def registration_ended(self):
        """Callback for the EndEvent."""
        plt.close()

    def registration_started(self):
        """Callback for the StartEvent."""
        self.metric_values = []
        self.resolution_iterations = []

    def registration_resolution_changed(self):
        """Callback for the MultiResolutionIterationEvent."""
        self.resolution_iterations.append(len(self.metric_values))

    def registration_iteration_ended(self):
        """Callback for the IterationEvent.

        Saves an image including the visualization of the registered images and the metric value plot.
        """

        self.metric_values.append(self.registration_method.GetMetricValue())
        # Plot the similarity metric values; resolution changes are marked with a blue star
        plt.plot(self.metric_values, 'r')
        plt.plot(self.resolution_iterations, [self.metric_values[index] for index in self.resolution_iterations], 'b*')
        plt.xlabel('Iteration', fontsize=12)
        plt.ylabel('Metric Value', fontsize=12)

        # todo(fabianbalsiger): format precision of legends
        # plt.axes().yaxis.set_major_formatter(matplotlib.ticker.StrMethodFormatter('{x:2f}'))
        # plt.axes().xaxis.set_major_formatter(matplotlib.ticker.StrMethodFormatter('{x:d}'))
        # _, ax = plt.subplots()
        # ax.yaxis.set_major_formatter(matplotlib.ticker.StrMethodFormatter('{x:f2}'))

        # todo(fabianbalsiger): add margin to the left side of the plot

        # Convert the plot to a SimpleITK image (works with the agg matplotlib backend, doesn't work
        # with the default - the relevant method is canvas_tostring_rgb())
        plt.gcf().canvas.draw()
        plot_data = np.fromstring(plt.gcf().canvas.tostring_rgb(), dtype=np.uint8, sep='')
        plot_data = plot_data.reshape(plt.gcf().canvas.get_width_height()[::-1] + (3,))
        plot_image = sitk.GetImageFromArray(plot_data, isVector=True)

        # Extract the central axial slice from the two volumes, compose it using the transformation and alpha blend it
        alpha = 0.7

        moving_transformed = sitk.Resample(self.moving_image, self.fixed_image, self.transform,
                                           sitk.sitkLinear, 0.0,
                                           self.moving_image.GetPixelIDValue())
        # Extract the plotting slice in xy and alpha blend them
        if self.fixed_image.GetDimension() == 3:
            slice_index = self.slice_no if self.slice_no != -1 else round((self.fixed_image.GetSize())[2] / 2)
            image_registration_overlay = (1.0 - alpha) * sitk.Normalize(self.fixed_image[:, :, slice_index]) + \
                       alpha * sitk.Normalize(moving_transformed[:, :, slice_index])
        else:
            image_registration_overlay = (1.0 - alpha) * sitk.Normalize(self.fixed_image) + \
                       alpha * sitk.Normalize(moving_transformed[:, :])

        combined_slices_image = sitk.ScalarToRGBColormap(image_registration_overlay)

        self._write_combined_image(combined_slices_image, plot_image,
                                   os.path.join(self.plot_dir,
                                                self.file_name_prefix + format(len(self.metric_values), '03d') + '.png')
                                   )

    @staticmethod
    def _write_combined_image(image1, image2, file_name):
        """Writes an image including the visualization of the registered images and the metric value plot."""
        combined_image = sitk.Image(
            (image1.GetWidth() + image2.GetWidth(), max(image1.GetHeight(), image2.GetHeight())),
            image1.GetPixelID(), image1.GetNumberOfComponentsPerPixel())

        image1_destination = [0, 0]
        image2_destination = [image1.GetWidth(), 0]

        if image1.GetHeight() > image2.GetHeight():
            image2_destination[1] = round((combined_image.GetHeight() - image2.GetHeight()) / 2)
        else:
            image1_destination[1] = round((combined_image.GetHeight() - image1.GetHeight()) / 2)

        combined_image = sitk.Paste(combined_image, image1, image1.GetSize(), (0, 0), image1_destination)
        combined_image = sitk.Paste(combined_image, image2, image2.GetSize(), (0, 0), image2_destination)
        sitk.WriteImage(combined_image, file_name)


class PlotOnResolutionChangeCallback(RegistrationCallback):
    """Represents a plotter for SimpleITK registrations.

    Saves the moving image on each resolution change and the registration end.
    """

    def __init__(self, plot_dir: str, file_name_prefix: str='') -> None:
        """
        Args:
            plot_dir (str): Path to the directory where to save the plots.
            file_name_prefix (str): The file name prefix for the plots.
        """
        super().__init__()
        self.plot_dir = plot_dir
        self.file_name_prefix = file_name_prefix

        self.resolution = 0

    def registration_ended(self):
        """Callback for the EndEvent."""
        self._write_image('end')

    def registration_started(self):
        """Callback for the StartEvent."""
        self.resolution = 0

    def registration_resolution_changed(self):
        """Callback for the MultiResolutionIterationEvent."""
        self._write_image('res' + str(self.resolution))
        self.resolution = self.resolution + 1

    def registration_iteration_ended(self):
        """Callback for the IterationEvent."""

    def _write_image(self, file_name_suffix: str):
        """Writes an image."""
        file_name = os.path.join(self.plot_dir, self.file_name_prefix + '_' + file_name_suffix + '.mha')

        moving_transformed = sitk.Resample(self.moving_image, self.fixed_image, self.transform,
                                           sitk.sitkLinear, 0.0, self.moving_image.GetPixelIDValue())
        sitk.WriteImage(moving_transformed, file_name)



import abc
import os

import h5py
import numpy as np

import pymia.data.indexexpression as expr
import pymia.data.definition as df


class Reader(metaclass=abc.ABCMeta):
    """Represents the abstract dataset reader."""

    def __init__(self, file_path: str) -> None:
        """Initializes a new instance.

        Args:
            file_path(str): The path to the dataset file.
        """
        super().__init__()
        self.file_path = file_path

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __del__(self):
        self.close()

    @abc.abstractmethod
    def get_subject_entries(self) -> list:
        """Get the dataset entries holding the subject's data.

        Returns:
            list: The list of subject entry strings.
        """
        pass

    @abc.abstractmethod
    def get_shape(self, entry: str) -> list:
        """Get the shape from an entry.

        Args:
            entry(str): The dataset entry.

        Returns:
            list: The shape of each dimension.
        """
        pass

    @abc.abstractmethod
    def get_subjects(self) -> list:
        """Get the subject names in the dataset.

        Returns:
            list: The list of subject names.
        """
        pass

    @abc.abstractmethod
    def read(self, entry: str, index: expr.IndexExpression=None):
        """Read a dataset entry.

        Args:
            entry(str): The dataset entry.
            index(expr.IndexExpression): The slicing expression.

        Returns:
            The read data.
        """
        pass

    @abc.abstractmethod
    def has(self, entry: str) -> bool:
        """Check whether a dataset entry exists.

        Args:
            entry(str): The dataset entry.

        Returns:
            bool: Whether the entry exists.
        """
        pass

    @abc.abstractmethod
    def open(self):
        """Open the reader."""
        pass

    @abc.abstractmethod
    def close(self):
        """Close the reader."""
        pass


class Hdf5Reader(Reader):
    """Represents the dataset reader for HDF5 files."""

    def __init__(self, file_path: str, category='images') -> None:
        """Initializes a new instance.

        Args:
            file_path(str): The path to the dataset file.
            category(str): The category of an entry that contains data of all subjects
        """
        super().__init__(file_path)
        self.h5 = None  # type: h5py.File
        self.category = category

    def get_subject_entries(self) -> list:
        group = df.DATA_PLACEHOLDER.format(self.category)
        return ['{}/{}'.format(group, k) for k in sorted(self.h5[group].keys())]

    def get_shape(self, entry: str) -> list:
        return self.h5[entry].shape

    def get_subjects(self) -> list:
        return self.read(df.SUBJECT)

    def read(self, entry: str, index: expr.IndexExpression=None):
        if index is None:
            data = self.h5[entry][()]  # need () instead of util.IndexExpression(None) [which is equal to slice(None)]
        else:
            data = self.h5[entry][index.expression]

        if isinstance(data, np.ndarray) and data.dtype == np.object:
            return data.tolist()
        # if h5py.check_dtype(vlen=self.h5[entry].dtype) == str and not isinstance(data, str):
        #     return data.tolist()
        return data

    def has(self, entry: str) -> bool:
        return entry in self.h5

    def open(self):
        self.h5 = h5py.File(self.file_path, mode='r', libver='latest')

    def close(self):
        if self.h5 is not None:
            self.h5.close()
            self.h5 = None


def get_reader(file_path: str, direct_open: bool=False) -> Reader:
    """ Get the dataset reader corresponding to the file extension.

    Args:
        file_path(str): The path to the dataset file.
        direct_open(bool): Whether the file should directly be opened.

    Returns:
        Reader: Reader corresponding to dataset file extension.
    """

    extension = os.path.splitext(file_path)[1]
    if extension not in reader_registry:
        raise ValueError('unknown dataset file extension "{}"'.format(extension))

    reader = reader_registry[extension](file_path)
    if direct_open:
        reader.open()
    return reader


reader_registry = {'.h5': Hdf5Reader, '.hdf5': Hdf5Reader}
